# -*- encoding: utf-8 -*-
"""
A collection of small helper functions for generating small pieces of datatable output in custom
methods on a view.

Because of uncontrolled external use of ``preload_record_data()`` by the view, each of these utility
functions allows for generic ``*args`` to pass through to the function, even though they aren't used
in any way.

"""

from functools import partial, wraps
import operator

from django import get_version
from django.db.models import Model
try:
    from django.forms.utils import flatatt
except ImportError:
    from django.forms.util import flatatt

import six

from .utils import resolve_orm_path, XEDITABLE_FIELD_TYPES

if [int(v) for v in get_version().split('.')[0:2]] >= [1, 5]:
    from django.utils.timezone import localtime
else:
    localtime = None


def keyed_helper(helper):
    """
    Decorator for helper functions that operate on direct values instead of model instances.

    A keyed helper is one that can be used normally in the view's own custom callbacks, but also
    supports direct access in the column declaration, such as in the example:

        datatable_options = {
            'columns': [
                ('Field Name', 'fieldname', make_boolean_checkmark(key=attrgetter('fieldname'))),
            ],
        }

    With the help of a ``sort``-style ``key`` argument, the helper can receive all the information
    it requires in advance, so that the view doesn't have to go through the trouble of declaring
    a custom callback method that simply returns the value of the ``make_boolean_checkmark()``
    helper.

    If the attribute being fetched is identical to the one pointed to in the column declaration,
    even the ``key`` argument can be omitted:

        ('Field Name', 'fieldname', make_boolean_checkmark)),

    """

    @wraps(helper)
    def wrapper(instance=None, key=None, attr=None, *args, **kwargs):
        if set((instance, key, attr)) == {None}:
            # helper was called in place with neither important arg
            raise ValueError("If called directly, helper function '%s' requires either a model"
                             " instance, or a 'key' or 'attr' keyword argument." % helper.__name__)

        if instance is not None:
            return helper(instance, *args, **kwargs)

        if key is None and attr is None:
            attr = 'self'

        if attr:
            if attr == 'self':
                key = lambda obj: obj
            else:
                key = operator.attrgetter(attr)

        # Helper is used directly in the columns declaration.  A new callable is
        # returned to take the place of a callback.
        @wraps(helper)
        def helper_wrapper(instance, *args, **kwargs):
            return helper(key(instance), *args, **kwargs)
        return helper_wrapper

    wrapper._is_wrapped = True
    return wrapper

@keyed_helper
def link_to_model(instance, text=None, *args, **kwargs):
    """
    Returns HTML in the form::

        <a href="{{ instance.get_absolute_url }}">{{ text }}</a>

    If ``text`` is provided and isn't empty, it will be used as the hyperlinked text.

    If ``text`` isn't available, then ``kwargs['rich_value']`` will be consulted instead.

    Failing those checks, the helper will fall back to simply using ``unicode(instance)`` as the
    link text.

    If the helper is called in place (rather than providing the helper reference directly), it can
    receive a special ``key`` argument, which is a mapping function that will receiving the instance
    (once it is available) and return some value from that instance.  That new value will be sent to
    the helper in the place of the instance.

    Examples::

        # Generate a simple href tag for instance.get_absolute_url()
        name = columns.TextColumn("Name", sources=['name'],
                                          processor=link_to_model)

        # Generate an href tag for instance.relatedobject.get_absolute_url()
        # Note that without the use of link_to_model(key=...), the object going into
        # the link_to_model helper would be the row instance, not the thing looked up by the
        # column's sources.
        name = columns.TextColumn("Name", sources=['relatedobject__name'],
                                          processor=link_to_model(key=getattr('relatedobject')))
    """
    if not text:
        text = kwargs.get('rich_value') or six.text_type(instance)
    return u"""<a href="{0}">{1}</a>""".format(instance.get_absolute_url(), text)


@keyed_helper
def make_boolean_checkmark(value, true_value="<div class='text-center'><span class='glyphicon glyphicon-ok' style='color:green'></span></div>", false_value="<div class='text-center'><span class='glyphicon glyphicon-remove' style='color:red'></span></div>", *args, **kwargs):
    """
    Returns a unicode ✔ or ✘, configurable by pre-calling the helper with ``true_value`` and/or
    ``false_value`` arguments, based on the incoming value.

    The value at ``kwargs['default_value']`` is checked to see if it casts to a boolean ``True`` or
    ``False``, and returns the appropriate representation.

    Examples::

        # A DateTimeField can be sent to the helper to detect whether
        # or not it is None, and have a checkmark reflect that.
        is_published = columns.DateTimeColumn("Published", sources=['published_date'],
                                              processor=make_boolean_checkmark)

        # Make the 'false_value' blank so that only True-like items have an icon
        is_published = columns.DateTimeColumn("Published", sources=['published_date'],
                                              processor=make_boolean_checkmark(false_value=""))

    """
    value = kwargs.get('default_value', value)
    if value:
        return true_value
    return false_value


def itemgetter(k, ellipsis=False, key=None):
    """
    Looks up ``k`` as an index of the column's value.

    If ``k`` is a ``slice`` type object, then ``ellipsis`` can be given as a string to use to
    indicate truncation.  Alternatively, ``ellipsis`` can be set to ``True`` to use a default
    ``'...'``.

    If a ``key`` is given, it may be a function which maps the target value to something else
    before the item lookup takes place.

    Examples::

        # Choose an item from a list source.
        winner = columns.TextColumn("Winner", sources=['get_rankings'],
                                    processor=itemgetter(0))

        # Take instance.description[:30] and append "..." to the end if truncation occurs.
        description = columns.TextColumn("Description", sources=['description'],
                                         processor=itemgetter(slice(None, 30), ellipsis=True))

    """
    def helper(instance, *args, **kwargs):
        default_value = kwargs.get('default_value')
        if default_value is None:
            default_value = instance
        value = default_value[k]
        if ellipsis and isinstance(k, slice) and isinstance(value, six.string_types) and \
                len(default_value) > len(value):
            if ellipsis is True:
                value += "..."
            else:
                value += ellipsis
        return value

    if key:
        helper = keyed_helper(helper)(key=key)
    return helper


def attrgetter(attr, key=None):
    """
    Looks up ``attr`` on the target value. If the result is a callable, it will be called in place
    without arguments.

    If a ``key`` is given, it may be a function which maps the target value to something else
    before the attribute lookup takes place.

    Examples::

        # Explicitly selecting the sources and then using a processor to allow the model
        # method to organize the data itself, you can still provide all the necessary
        # ORM hints to the column.
        # This is definitely superior to having sources=['get_address'].
        address = columns.TextColumn("Address", sources=['street', 'city', 'state', 'zip'],
                                     processor=attrgetter('get_address'))

    """
    def helper(instance, *args, **kwargs):
        value = instance
        for bit in attr.split('.'):
            value = getattr(value, bit)
            if callable(value):
                value = value()
        return value

    if key:
        helper = keyed_helper(helper)(key=key)
    return helper


def format_date(format_string, localize=False, key=None):
    """
    A pre-called helper to supply a date format string ahead of time, so that it can apply to each
    date or datetime that this column represents.  With Django >= 1.5, the ``localize=True`` keyword
    argument can be given, or else can be supplied in the column's own declaration for the same
    effect.  (The date and datetime columns explicitly forward their ``localize`` setting to all
    helpers.)

    If the ``key`` argument is given, it may be a function which maps the target value to something
    else before the date formatting takes place.
    """

    if localize is not False and localtime is None:
        raise Exception("Cannot use format_date argument 'localize' with Django < 1.5")

    def helper(value, *args, **kwargs):
        inner_localize = kwargs.get('localize', localize)
        if inner_localize is not False and localtime is None:
            raise Exception("Cannot use format_date argument 'localize' with Django < 1.5")

        if key:
            value = key(value)
        else:
            value = kwargs.get('default_value', value)
        if not value:  # Empty or missing default_value
            return ""
        if localize:
            value = localtime(value)
        return value.strftime(format_string)

    if key:
        return keyed_helper(helper)(key=key)
    return helper


def format(format_string, cast=lambda x: x):
    """
    A pre-called helper to supply a modern string format (the kind with {} instead of %s), so that
    it can apply to each value in the column as it is rendered.  This can be useful for string
    padding like leading zeroes, or rounding floating point numbers to a certain number of decimal
    places, etc.

    If given, the ``cast`` argument should be a mapping function that coerces the input to whatever
    type is required for the string formatting to work.  Trying to push string data into a float
    format will raise an exception, for example, so the ``float`` type itself could be given as
    the ``cast`` function.

    Examples::
    
        # Perform some 0 padding
        item_number = columns.FloatColumn("Item No.", sources=['number'],
                                          processor=format("{:03d}))

        # Force a string column value to coerce to float and round to 2 decimal places
        rating = columns.TextColumn("Rating", sources=['avg_rating'],
                                    processor=format("{:.2f}", cast=float))

    """

    def helper(instance, *args, **kwargs):
        value = kwargs.get('default_value')
        if value is None:
            value = instance
        value = cast(value)
        return format_string.format(value, obj=instance)
    return helper


def make_xeditable(instance=None, extra_attrs=[], *args, **kwargs):
    """
    Converts the contents of the column into an ``<a>`` tag with the required DOM attributes to
    power the X-Editable UI.

    The following keyword arguments are all optional, but may be provided when pre-calling the
    helper, to customize the output of the helper once it is run per object record:

        * ``type`` - Defaults to the basic type of the HTML input ("text", "number", "datetime")
        * ``title`` - Defaults to an empty string, controls the HTML "title" attribute.
        * ``placeholder`` - Defaults to whatever "title" is, controls the HTML "placeholder" attribute.
        * ``url`` - Defaults to the ``request.path`` of the view, which will automatically serve the X-Editable interface as long as it inherits from ``XEditableDatatableView``.
        * ``source`` - Defaults to the ``request.path`` of the view, which will automatically serve X-Editable requests for ``choices`` data about a field.

    Supplying a list of names via ``extra_attrs`` will enable arbitrary other keyword arguments to
    be rendered in the HTML as attribute as well.  ``extra_attrs`` serves as a whitelist of extra
    names so that unintended kwargs don't get rendered without your permission.
    """

    if instance is None:
        # Preloading kwargs into the helper for deferred execution
        helper = partial(make_xeditable, extra_attrs=extra_attrs, *args, **kwargs)
        return helper

    # Immediate finalization, return the xeditable structure
    data = kwargs.get('default_value', instance)
    rich_data = kwargs.get('rich_value', data)

    # Compile values to appear as "data-*" attributes on the anchor tag
    default_attr_names = ['pk', 'type', 'url', 'source', 'title', 'placeholder']
    valid_attr_names = set(default_attr_names + list(extra_attrs))
    attrs = {}
    for k, v in kwargs.items():
        if k in valid_attr_names:
            if k.startswith('data_'):
                k = k[5:]
            attrs['data-{0}'.format(k)] = v

    attrs['data-xeditable'] = "xeditable"

    # Assign default values where they are not provided

    field_name = kwargs['field_name']  # sent as a default kwarg to helpers
    if isinstance(field_name, (tuple, list)):
        # Legacy syntax
        field_name = field_name[1]
        if isinstance(field_name, (tuple, list)):
            raise ValueError("'make_xeditable' helper needs a single-field data column,"
                             " not {0!r}".format(field_name))
    attrs['data-name'] = field_name

    if isinstance(rich_data, Model):
        attrs['data-value'] = rich_data.pk
    else:
        attrs['data-value'] = rich_data
    attrs['value'] = attrs['data-value']
    if 'data-pk' not in attrs:
        attrs['data-pk'] = instance.pk

    if 'data-url' not in attrs:
        # Look for a backup data-url
        provider_name = 'get_update_url'
        url_provider = getattr(kwargs.get('view'), provider_name, None)
        if not url_provider:
            url_provider = getattr(instance, provider_name, None)
            if not url_provider and 'view' in kwargs:
                url_provider = lambda field_name: kwargs['view'].request.path
            else:
                raise ValueError("'make_xeditable' cannot determine a value for 'url'.")
        if url_provider:
            attrs['data-url'] = url_provider(field_name=field_name)

    if 'data-placeholder' not in attrs:
        attrs['data-placeholder'] = attrs.get('data-title', "")

    if 'data-type' not in attrs:
        if hasattr(instance, '_meta'):
            # Try to fetch a reasonable type from the field's class
            if field_name == 'pk':  # special field name not in Model._meta.fields
                field = instance._meta.pk
            else:
                field = resolve_orm_path(instance, field_name)

            if field.choices:
                field_type = 'select'
            else:
                field_type = XEDITABLE_FIELD_TYPES.get(field.get_internal_type(), 'text')
        else:
            field_type = 'text'
        attrs['data-type'] = field_type

    # type=select elements need to fetch their valid choice options from an AJAX endpoint.
    # Register the view for this lookup.
    if attrs['data-type'] in ('select', 'select2'):
        if 'data-source' not in attrs:
            if 'view' in kwargs:
                attrs['data-source'] = "{url}?{field_param}={fieldname}".format(**{
                    'url': kwargs['view'].request.path,
                    'field_param': kwargs['view'].xeditable_fieldname_param,
                    'fieldname': field_name,
                })
                if attrs['data-type'] == 'select2':
                    attrs['data-source'] += '&select2=true'
            else:
                raise ValueError("'make_xeditable' cannot determine a value for 'source'.")

        # Choice fields will want to display their readable label instead of db data
        data = getattr(instance, 'get_{0}_display'.format(field_name), lambda: data)()
    if data=='True':
        attrs['value'] = 1
        data = u"""<div style="text-align:center"><input {attrs} type="checkbox" checked class="checkbox-active" style="top:.8rem;width:1.55rem;height:1.55rem;"></input></div>""".format(attrs=flatatt(attrs))
    elif data=='False':
        attrs['value'] = 0
        data = u"""<div style="text-align:center"><input {attrs} type="checkbox" class="checkbox-active" style="top:.8rem;width:1.55rem;height:1.55rem;"></input></div>""".format(attrs=flatatt(attrs))
    elif ((field_name == 'model') | (field_name=='target_exposure')):
        data = u"""<input type="number" {attrs} min="1" max="100" size="5" style="max-width: 8em" class="numberinput form-control editor-active" required="" {data}></input>""".format(attrs=flatatt(attrs), data=data)
    elif ((field_name == 'Team_Players') | (field_name == 'Opposing_Players')):
        data = u"""<input type="number" {attrs} min="1" max="{limit}" size="5" step="1" style="max-width: 8em" class="numberinput form-control editor-active" required="" {data}></input>""".format(attrs=flatatt(attrs), data=data, limit = 3 if field_name=='Team_Players' else 2)
    else:
        #data = u"""<input type="number" {attrs} min="1" max="150" size="5" style="max-width: 8em" class="numberinput form-control" required="" id="id_Lineup_Count"> href="#"{attrs}>{data}</a>""".format(attrs=flatatt(attrs), data=data)
        data = u"""<a href="#"{attrs}>{data}</a>""".format(attrs=flatatt(attrs), data=data)
    return data


def make_processor(func, arg=None):
    """
    A pre-called processor that wraps the execution of the target callable ``func``.

    This is useful for when ``func`` is a third party mapping function that can take your column's
    value and return an expected result, but doesn't understand all of the extra kwargs that get
    sent to processor callbacks.  Because this helper proxies access to ``func``, it can hold back
    the extra kwargs for a successful call.

    ``func`` will be called once per object record, a single positional argument being the column
    data retrieved via the column's :py:attr:`~datatableview.columns.Column.sources`

    An optional ``arg`` may be given, which will be forwarded as a second positional argument to
    ``func``.  This was originally intended to simplify using Django template filter functions as
    ``func``.  If you need to sent more arguments, consider wrapping your ``func`` in a
    ``functools.partial``, and use that as ``func`` instead.
    """
    def helper(instance, *args, **kwargs):
        value = kwargs.get('default_value')
        if value is None:
            value = instance
        if arg is not None:
            extra_arg = [arg]
        else:
            extra_arg = []
        return func(value, *extra_arg)
    return helper

through_filter = make_processor



# -*- encoding: utf-8 -*-

import re

try:
    from django.urls import url
except ImportError:
    from django.conf.urls import url

from . import views

urls = []
for attr in dir(views):
    View = getattr(views, attr)
    try:
        is_demo = issubclass(View, views.DemoMixin) and View is not views.DemoMixin
    except TypeError:
        continue
    if is_demo:
        name = re.sub(r'([a-z]|[A-Z]+)(?=[A-Z])', r'\1-', attr).lower()
        name = name.replace("-datatable-view", "")
        urls.append(url(r'^{name}/$'.format(name=name), View.as_view(), name=name))

urlpatterns = [
    url(r'^$', views.IndexView.as_view(), name="index"),
    url(r'^reset/$', views.ResetView.as_view()),
    url(r'^migration-guide/$', views.MigrationGuideView.as_view(), name="migration-guide"),
    url(r'^column-formats/$', views.ValidColumnFormatsView.as_view(), name="column-formats"),
    url(r'^javascript-initialization/$', views.JavascriptInitializationView.as_view(), name="js-init"),
    url(r'^satellite/$', views.SatelliteDatatableView.as_view(), name="satellite"),
] + urls



# -*- encoding: utf-8 -*-

import re
import copy
import operator
from collections import OrderedDict
try:
    from functools import reduce
except ImportError:
    pass

from django.db.models.fields import FieldDoesNotExist
from django.template.loader import render_to_string
from django.db.models import QuerySet
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_unicode as force_text
try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError:
    from .compat import python_2_unicode_compatible

import six


from .exceptions import ColumnError, SkipRecord
from .columns import (Column, TextColumn, DateColumn, DateTimeColumn, BooleanColumn, IntegerColumn,
                      FloatColumn, DisplayColumn, CompoundColumn, get_column_for_modelfield)
from .utils import (OPTION_NAME_MAP, MINIMUM_PAGE_LENGTH, contains_plural_field, split_terms,
                    resolve_orm_path)
from .cache import DEFAULT_CACHE_TYPE, cache_types, get_cache_key, cache_data, get_cached_data

def pretty_name(name):
    if not name:
        return ''
    return name[0].capitalize() + name[1:]


# Borrowed from the Django forms implementation
def columns_for_model(model, fields=None, exclude=None, labels=None, processors=None,
                      unsortable=None, hidden=None):
    field_list = []
    opts = model._meta
    for f in sorted(opts.fields):
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue

        column_class = get_column_for_modelfield(f)
        if column_class is None:
            raise ColumnError("Unhandled model field %r." % (f,))
        if labels and f.name in labels:
            label = labels[f.name]
        else:
            label = f.verbose_name
        if processors and f.name in processors:
            processor = processors[f.name]
        else:
            processor = None
        if unsortable and f.name in unsortable:
            sortable = False
        else:
            sortable = True
        if hidden and f.name in hidden:
            visible = False
        else:
            visible = True
        label = (labels or {}).get(f.name, pretty_name(f.verbose_name))
        column = column_class(sources=[f.name], label=label, processor=processor, sortable=sortable,
                              visible=visible)
        column.name = f.name
        field_list.append((f.name, column))

    field_dict = OrderedDict(field_list)
    if fields:
        field_dict = OrderedDict(
            [(f, field_dict.get(f)) for f in fields
                if (not exclude) or (exclude and f not in exclude)]
        )
    return field_dict

# Borrowed from the Django forms implementation
def get_declared_columns(bases, attrs, with_base_columns=True):
    """
    Create a list of form field instances from the passed in 'attrs', plus any
    similar fields on the base classes (in 'bases'). This is used by both the
    Form and ModelForm metclasses.

    If 'with_base_columns' is True, all fields from the bases are used.
    Otherwise, only fields in the 'declared_fields' attribute on the bases are
    used. The distinction is useful in ModelForm subclassing.
    Also integrates any additional media definitions
    """
    local_columns = [
        (column_name, attrs.pop(column_name)) \
                for column_name, obj in list(six.iteritems(attrs)) \
                if isinstance(obj, Column)
    ]
    local_columns.sort(key=lambda x: x[1].creation_counter)

    # If this class is subclassing another Form, add that Form's columns.
    # Note that we loop over the bases in *reverse*. This is necessary in
    # order to preserve the correct order of columns.
    if with_base_columns:
        for base in bases[::-1]:
            if hasattr(base, 'base_columns'):
                local_columns = list(six.iteritems(base.base_columns)) + local_columns
    else:
        for base in bases[::-1]:
            if hasattr(base, 'declared_columns'):
                local_columns = list(six.iteritems(base.declared_columns)) + local_columns

    return OrderedDict(local_columns)

class DatatableOptions(object):
    """
    Contains declarable options for a datatable, some of which can be manipuated by subsequent
    requests by the user.
    """
    def __init__(self, options=None):
        # Non-mutable; server's declared preference is final
        self.id = getattr(options, 'id', '')
        self.model = getattr(options, 'model', None)
        self.columns = getattr(options, 'columns', None)  # table headers
        self.exclude = getattr(options, 'exclude', None)
        self.search_fields = getattr(options, 'search_fields', None)  # extra searchable ORM fields
        self.unsortable_columns = getattr(options, 'unsortable_columns', None)
        self.hidden_columns = getattr(options, 'hidden_columns', None)  # generated, but hidden
        self.labels = getattr(options, 'labels', None)
        self.processors = getattr(options, 'processors', None)
        self.request_method = getattr(options, 'request_method', 'GET')
        self.structure_template = getattr(options, 'structure_template', "datatableview/default_structure.html")
        self.footer = getattr(options, 'footer', False)
        self.result_counter_id = getattr(options, 'result_counter_id', 'id_count')

        # Non-mutable; server behavior customization
        self.cache_type = getattr(options, 'cache_type', cache_types.NONE)
        self.cache_queryset_count = getattr(options, 'cache_queryset_count', False)

        # Mutable by the request
        self.ordering = getattr(options, 'ordering', None)  # override to Model._meta.ordering
        self.page_length = getattr(options, 'page_length', 25)  # length of a single result page


default_options = DatatableOptions()

class DatatableMetaclass(type):
    """
    Each declared Datatable object inspects its declared "fields" in order to facilitate an
    inheritance system resembling the django.forms system.  Except for our custom Meta options that
    offer field options ('labels', 'processors', etc), this code is essentially a clone of the
    django.forms strategy.
    """

    def __new__(cls, name, bases, attrs):
        declared_columns = get_declared_columns(bases, attrs, with_base_columns=False)
        new_class = super(DatatableMetaclass, cls).__new__(cls, name, bases, attrs)

        opts = new_class._meta = new_class.options_class(getattr(new_class, 'Meta', None))
        if opts.model:
            columns = columns_for_model(opts.model, opts.columns, opts.exclude, opts.labels,
                                        opts.processors, opts.unsortable_columns, opts.hidden_columns)
            none_model_columns = [k for k, v in six.iteritems(columns) if not v]
            missing_columns = set(none_model_columns) - set(declared_columns.keys())

            for name, column in declared_columns.items():
                column.name = name
                # if not column.sources:
                #     column.sources = [name]
                if not column.label:
                    try:
                        field = resolve_orm_path(opts.model, name)
                    except FieldDoesNotExist:
                        label = name
                    else:
                        label = field.verbose_name
                    column.label = pretty_name(label)

            # Normalize declared 'search_fields' to Column instances
            if isinstance(opts.search_fields, dict):
                # Turn a dictionary of {name: ColumnClass} to just a list of classes.
                # If only the column class reference is given instead of an instance, instantiate
                # the object first.
                search_fields = []
                for name, column in opts.search_fields.items():
                    if callable(column):
                        column = column(sources=[name])
                    search_fields.append(column)
                opts.search_fields = search_fields
            elif opts.search_fields is None:
                opts.search_fields = []
            else:
                opts.search_fields = list(opts.search_fields)
            for i, column in enumerate(opts.search_fields):
                # Build a column object
                if isinstance(column, six.string_types):
                    name = column
                    field = resolve_orm_path(opts.model, name)
                    column = get_column_for_modelfield(field)
                    opts.search_fields[i] = column(sources=[name])

            columns.update(declared_columns)
        else:
            columns = declared_columns
            missing_columns = []

        new_class.declared_columns = declared_columns
        new_class.base_columns = columns
        new_class.missing_columns = missing_columns
        return new_class


@python_2_unicode_compatible
class Datatable(six.with_metaclass(DatatableMetaclass)):
    """
    Declaration container for a clientside datatable, containing an optional Meta inner class,
    class-level field declarations, and callbacks for filtering and post-processing values requested
    by the client.

    Client options sent over AJAX GET parameters will override the settings given in the inner Meta
    class.

    This object can be sent to a template context and rendered there in order to generate an
    annotated HTML frame for the javascript to initialize.
    """

    options_class = DatatableOptions

    def __init__(self, object_list, url, view=None, callback_target=None, model=None,
                 query_config=None, force_distinct=True, **kwargs):
        self.object_list = object_list
        self.url = url
        self.view = view
        self.forward_callback_target = callback_target
        self.model = model or self._meta.model
        if self.model is None and hasattr(object_list, 'model'):
            self.model = object_list.model
        if query_config is None:
            query_config = {}
        self.query_config = query_config

        self.columns = copy.deepcopy(self.base_columns)

        self._force_distinct = force_distinct
        self.total_initial_record_count = None
        self.unpaged_record_count = None

    def configure(self):
        """
        Combines (in order) the declared/inherited inner Meta, any view options, and finally any
        valid AJAX GET parameters from client modifications to the data they see.
        """

        if hasattr(self, '_configured'):
            return

        self.resolve_virtual_columns(*tuple(self.missing_columns))

        self.config = self.normalize_config(self._meta.__dict__, self.query_config)

        self.config['column_searches'] = {}
        for i, name in enumerate(self.columns.keys()):
            column_search = self.query_config.get(OPTION_NAME_MAP['search_column'] % i, None)
            if column_search:
                self.config['column_searches'][name] = column_search

        column_order = list(self.columns.keys())
        if self.config['ordering']:
            for i, name in enumerate(self.config['ordering']):
                column_name = name.lstrip('-+')
                try:
                    index = column_order.index(column_name)
                except ValueError:
                    # It is important to ignore a bad ordering name, since the model.Meta may
                    # specify a field name that is not present on the datatable columns list.
                    continue
                self.columns[column_name].sort_priority = i
                self.columns[column_name].sort_direction = 'desc' if name[0] == '-' else 'asc'
                self.columns[column_name].index = index

        for name in self.config['hidden_columns']:
            self.columns[name].visible = False

        self._configured = True

    # Client request configuration mergers
    def normalize_config(self, config, query_config):
        """
        Merge the declared configuration with whatever valid query parameters are found from the
        client's AJAX request.
        """

        # Core options, not modifiable by client updates
        if config['hidden_columns'] is None:
            config['hidden_columns'] = []
        if config['search_fields'] is None:
            config['search_fields'] = []
        if config['unsortable_columns'] is None:
            config['unsortable_columns'] = []

        config['search'] = self.normalize_config_search(config, query_config)
        config['start_offset'] = self.normalize_config_start_offset(config, query_config)
        config['page_length'] = self.normalize_config_page_length(config, query_config)
        config['ordering'] = self.normalize_config_ordering(config, query_config)
        self._ordering_columns = self.ensure_ordering_columns(config['ordering'])

        return config

    def normalize_config_search(self, config, query_config):
        terms_string = query_config.get(OPTION_NAME_MAP['search'], '').strip()
        return set(split_terms(terms_string))

    def normalize_config_start_offset(self, config, query_config):
        try:
            start_offset = query_config.get(OPTION_NAME_MAP['start_offset'], 0)
            start_offset = int(start_offset)
        except ValueError:
            start_offset = 0
        else:
            if start_offset < 0:
                start_offset = 0
        return start_offset

    def normalize_config_page_length(self, config, query_config):
        try:
            page_length = query_config.get(OPTION_NAME_MAP['page_length'], config['page_length'])
            page_length = int(page_length)
        except ValueError:
            page_length = config['page_length']
        else:
            if page_length == -1:  # dataTables' way of asking for all items, no pagination
                pass
            elif page_length < MINIMUM_PAGE_LENGTH:
                page_length = MINIMUM_PAGE_LENGTH
        return page_length

    def normalize_config_ordering(self, config, query_config):
        default_ordering = config['ordering']
        if default_ordering is None and config['model']:
            default_ordering = config['model']._meta.ordering

        sort_declarations = [k for k in query_config if re.match(r'^order\[\d+\]\[column\]$', k)]

        # Default sorting from view or model definition
        if len(sort_declarations) == 0:
            return default_ordering

        ordering = []
        columns_list = list(self.columns.values())

        for sort_queue_i in range(len(columns_list)):
            try:
                column_index = int(query_config.get(OPTION_NAME_MAP['sort_column'] % sort_queue_i, ''))
            except ValueError:
                continue

            column = columns_list[column_index]

            # Reject requests for unsortable columns
            if column.name in config['unsortable_columns']:
                continue

            sort_direction = query_config.get(OPTION_NAME_MAP['sort_column_direction'] % sort_queue_i, None)

            if sort_direction == 'asc':
                sort_modifier = ''
            elif sort_direction == 'desc':
                sort_modifier = '-'
            else:
                # Aggressively skip invalid specification
                continue

            ordering.append('%s%s' % (sort_modifier, column.name))

        if not ordering:
            return default_ordering
        return ordering

    def ensure_ordering_columns(self, ordering_names):
        if ordering_names is None:
            return {}

        # Normalize declared 'ordering' to Column instances
        ordering_columns = {}
        for i, name in enumerate(ordering_names):
            if name[0] in '+-':
                name = name[1:]

            if name not in self.columns:
                field = resolve_orm_path(self.model, name)
                column = get_column_for_modelfield(field)
                ordering_columns[name] = column(sources=[name])

        return ordering_columns

    def resolve_virtual_columns(self, *names):
        """
        Called with ``*args`` from the Meta.columns declaration that don't match the model's known
        fields.  This method can inspect these names and decide what to do with them in special
        scenarios, but by default, they are simply raised in an exception to notify the developer of
        an apparent configuration error.
        """
        if names:
            raise ColumnError("Unknown column name(s): %r" % (names,))

    # Reflection methods for wrapped columns
    def get_ordering_splits(self):
        """
        Returns a 2-tuple of database-sortable and non-database-sortable column names.  The first
        list ends when the first non-db column is found.  It is therefore possible that
        ``virtual_fields`` contains subsequent real db-backed fields, but because arrangement of the
        ordering fields matters, we can't respect those until manual ordering has been done on the
        intervening non-db fields.
        """
        if self.config['ordering'] is None:
            return [], []

        db_fields = []
        virtual_fields = []
        i = 0
        for i, name in enumerate(self.config['ordering']):
            if name[0] in '+-':
                name = name[1:]
            if name in self.columns:
                column = self.columns[name]
            else:
                column = self._ordering_columns[name]
            if not column.get_db_sources(self.model):
                break
        else:
            i = len(self.config['ordering'])
        return self.config['ordering'][:i], self.config['ordering'][i:]

    def get_db_splits(self):
        """ Legacy utility for fetching the database columns and non-database columns. """
        db_fields = []
        virtual_fields = []
        for name, column in self.columns.items():
            if column.get_db_sources():
                db_fields.append(name)
            else:
                virtual_fields.append(name)
        return db_fields, virtual_fields

    # Data retrieval
    def will_load_from_cache(self, **kwargs):
        """
        Returns a hint for external code concerning the presence of cache data for the given kwargs.

        See :py:meth:`.get_cache_key_kwargs` for information concerning the kwargs you must send for
        this hint to be accurate.
        """
        cached_data = self.get_cached_data(datatable_class=self.__class__, **kwargs)
        return (type(cached_data) is not type(None))

    def get_cache_key_kwargs(self, view=None, user=None, **kwargs):
        """
        Returns the dictionary of kwargs that will be sent to :py:meth:`.get_cache_key` in order to
        generate a deterministic cache key.

        ``datatable_class``, ``view``, and ``user`` are returned by default, the user being looked
        up on the view's ``request`` attribute.

        Override this classmethod in order to add or remove items from the returned dictionary if
        you need a more specific or less specific cache key.
        """
        # Try to get user information if 'user' param is missing
        if hasattr(view, 'request') and not user:
            user = view.request.user

        kwargs.update({
            'datatable_class': self.__class__,
            'view': view,
            'user': user,
        })

        return kwargs

    def get_cache_key(self, **kwargs):
        """
        Returns the full cache key used for object_list data handled by this datatable class.
        ``settings.DATATABLEVIEW_CACHE_PREFIX`` will be prepended to this value.

        The kwargs sent guarantee a deterministic cache key between requests.

        ``view`` and ``user`` are special kwargs that the caching system provides by default.  The
        view instance is inspected for its ``__module__.__name__`` string, and the user for its
        ``pk``.

        All other kwargs are hashed and appended to the cache key.
        """
        return get_cache_key(**kwargs)

    def get_cached_data(self, **kwargs):
        """ Returns object_list data cached for the given kwargs. """
        return get_cached_data(self, **kwargs)

    def cache_data(self, data, **kwargs):
        """ Caches object_list data for the given kwargs. """
        cache_data(self, data=data, **kwargs)

    def get_object_list(self):
        """
        Returns a cached object list if configured and available.  When no caching strategy is
        enabled or if the cached item is expired, the original ``object_list`` is returned.
        """

        # Initial object_list from constructor, before filtering or ordering.
        object_list = self.object_list

        # Consult cache, if enabled
        cache_type = self.config['cache_type']
        if cache_type == cache_types.DEFAULT:
            cache_type = DEFAULT_CACHE_TYPE

        if cache_type:
            cache_kwargs = self.get_cache_key_kwargs(view=self.view)
            cached_data = self.get_cached_data(**cache_kwargs)

            # If no cache is available, simplify and store the original object_list
            if cached_data is None:
                cached_data = self.prepare_object_list_for_cache(cache_type, object_list)
                self.cache_data(cached_data, **cache_kwargs)

            object_list = self.expand_object_list_from_cache(cache_type, cached_data)

        return object_list

    def prepare_object_list_for_cache(self, cache_type, object_list):
        """
        Pre-caching hook that must prepare ``object_list`` for the cache using the strategy
        indicated by ``cache_type``, which is the table's ``Meta``
        :py:attr:`~datatableview.datatables.Meta.cache_type` value.

        When ``cache_type`` is ``SIMPLE``, the ``object_list`` is returned unmodified.

        When ``PK_LIST`` is used, ``object_list`` is queried for the list of ``pk`` values and those
        are returned instead.
        """
        data = object_list

        # Create the simplest reproducable query for repeated operations between requests
        # Note that 'queryset' cache_type is unhandled so that it passes straight through.
        if cache_type == cache_types.PK_LIST:
            model = object_list.model
            data = tuple(object_list.values_list('pk', flat=True))

        # Objects in some other type of data structure should be pickable for cache backend
        return data

    def expand_object_list_from_cache(self, cache_type, cached_data):
        """
        Deserializes the ``cached_data`` fetched from the caching backend, according to the
        ``cache_type`` strategy that was used to originally store it.

        When ``cache_type`` is ``SIMPLE``, the ``cached_data`` is returned unmodified, since the
        ``object_list`` went into the cache unmodified.

        When ``PK_LIST`` is used, ``cached_data`` is treated as a list of ``pk`` values and is used
        to filter the model's default queryset to just those objects.
        """
        if cache_type == cache_types.PK_LIST:
            # Convert pk list back into queryset
            data = self.model.objects.filter(pk__in=cached_data)
        else:
            # Straight passthrough of cached items
            data = cached_data
        return data

    def _get_current_page(self):
        """
        If page_length is specified in the options or AJAX request, the result list is shortened to
        the correct offset and length.  Paged or not, the finalized object_list is then returned.
        """

        # Narrow the results to the appropriate page length for serialization
        if self.config['page_length'] != -1:
            i_begin = self.config['start_offset']
            i_end = self.config['start_offset'] + self.config['page_length']
            object_list = self._records[i_begin:i_end]
        else:
            object_list = self._records

        return object_list

    def get_records(self):
        """
        Calls :py:meth:`.populate_records` to apply searches and sorting to the object list, then
        extracts the applicate page of results, calling :py:meth:`.get_record_data` for each result
        in the page.

        Returns the final list of processed results.
        """
        if not hasattr(self, '_records'):
            self.populate_records()

        page_data = []
        for obj in self._get_current_page():
            try:
                record_data = self.get_record_data(obj)
            except SkipRecord:
                pass
            else:
                page_data.append(record_data)
        return page_data

    def populate_records(self):
        """
        Searches and sorts the original object list.  Even though these operations do not themselves
        generate queries, the original object list will be counted and the filtered object list will
        also be counted, meaning that this is the method is guaranteed to run queries.

        No paging will take place at this stage!
        """
        if not hasattr(self, 'config'):
            self.configure()

        self._records = None
        base_objects = self.get_object_list()
        filtered_objects = self.search(base_objects)
        filtered_objects = self.sort(filtered_objects)
        self._records = filtered_objects

        num_total, num_filtered = self.count_objects(base_objects, filtered_objects)
        self.total_initial_record_count = num_total
        self.unpaged_record_count = num_filtered

    def count_objects(self, base_objects, filtered_objects):
        """
        Calculates object totals for datatable footer.  Returns a 2-tuple of counts for,
        respectively, the total number of objects and the filtered number of objects.

        Up to two ``COUNT`` queries may be issued.  If you already have heavy backend queries, this
        might add significant overhead to every ajax fetch, such as keystroke filters.

        If ``Meta.cache_type`` is configured and ``Meta.cache_queryset_count`` is set to True, the
        resulting counts will be stored in the caching backend.
        """

        num_total = None
        num_filtered = None

        if isinstance(base_objects, QuerySet):
            if self.config['cache_queryset_count']:
                cache_kwargs = self.get_cache_key_kwargs(view=self.view, __num_total='__num_total')
                num_total = self.get_cached_data(**cache_kwargs)

            if num_total is None:
                num_total = base_objects.count()
                if self.config['cache_queryset_count']:
                    self.cache_data(num_total, **cache_kwargs)
        else:
            num_total = len(base_objects)

        if len(self.config['search']) > 0 or len(self.config['column_searches']) > 0:
            if isinstance(filtered_objects, QuerySet):
                num_filtered = filtered_objects.count()
            else:
                num_filtered = len(filtered_objects)
        else:
            num_filtered = num_total

        return num_total, num_filtered

    def search(self, queryset):
        """ Performs db-only queryset searches. """

        table_queries = []

        searches = {}

        # Add per-column searches where necessary
        for name, term in self.config['column_searches'].items():
            for term in set(split_terms(term)):
                columns = searches.setdefault(term, {})
                columns[name] = self.columns[name]

        # Global search terms apply to all columns
        for term in self.config['search']:
            # NOTE: Allow global terms to overwrite identical queries that were single-column
            searches[term] = self.columns.copy()
            searches[term].update({None: column for column in self.config['search_fields']})

        for term in searches.keys():
            term_queries = []
            for name, column in searches[term].items():
                if name is None:  # config.search_fields items
                    search_f = self._search_column
                else:
                    search_f = getattr(self, 'search_%s' % (name,), self._search_column)
                q = search_f(column, term)
                if q is not None:
                    term_queries.append(q)
            if term_queries:
                table_queries.append(reduce(operator.or_, term_queries))

        if table_queries:
            q = reduce(operator.and_, table_queries)
            queryset = queryset.filter(q)

        return queryset.distinct()

    def _search_column(self, column, terms):
        """ Requests search queries to be performed against the target column.  """
        return column.search(self.model, terms)

    def sort(self, queryset):
        """
        Performs db-only queryset sorts, then applies manual sorts if required.
        """
        fields = []
        db, virtual = self.get_ordering_splits()
        for name in db:
            sort_direction = ''
            if name[0] in '+-':
                sort_direction = name[0]
                if sort_direction == '+':
                    sort_direction = ''
                name = name[1:]
            if name in self.columns:
                column = self.columns[name]
            else:
                column = self._ordering_columns[name]
            sources = column.get_sort_fields(self.model)
            if sources:
                fields.extend([(sort_direction + source) for source in sources])

        object_list = queryset.order_by(*fields)

        # When sorting a plural relationship field, we get duplicate rows for each item on the other
        # end of that relationship, which can't be removed with a call to distinct().
        if self._force_distinct and contains_plural_field(self.model, fields):
            object_list = self.force_distinct(object_list)

        if virtual:
            # Have to sort the whole queryset by hand!
            object_list = list(object_list)

            def flatten(value):
                if isinstance(value, (list, tuple)):
                    return flatten(value[0])
                return value

            for name in virtual[::-1]:  # stable sorting, top priority sort comes last
                reverse = False
                if name[0] in '+-':
                    reverse = (name[0] == '-')
                    name = name[1:]
                column = self.columns[name]
                object_list.sort(key=lambda o: flatten(column.value(o)[0]), reverse=reverse)

        return object_list

    def force_distinct(self, object_list):
        seen = set()
        def is_unseen(obj):
            if obj.pk in seen:
                return False
            seen.add(obj.pk)
            return True
        return tuple(obj for obj in object_list if is_unseen(obj))

    # Per-record callbacks
    def preload_record_data(self, obj):
        """
        An empty hook for doing something with a result ``obj`` before column lookups are called
        against the object.  The dict of items returned will be passed as keyword arguments to any
        available column ``processor`` callbacks.

        Use this to look up expensive data once per record so that it can be shared between column
        processors.

        By default, this method also inspects the originating view for a method of the same name,
        giving it an opportunity to contribute to the preloaded data.
        """

        kwargs = {}
        if self.forward_callback_target and \
                hasattr(self.forward_callback_target, 'preload_record_data'):
            kwargs.update(self.forward_callback_target.preload_record_data(obj))
        return kwargs

    def get_object_pk(self, obj):
        """ Returns the object's ``pk`` value. """
        return obj.pk

    def get_extra_record_data(self, obj):
        """ Returns a dictionary of JSON-friendly data sent to the client as ``"DT_RowData"``. """
        data = {}
        if self.forward_callback_target and \
                hasattr(self.forward_callback_target, 'get_extra_record_data'):
            data.update(self.forward_callback_target.get_extra_record_data(obj))
        return data

    def get_record_data(self, obj):
        """
        Returns a dict of column data that will be given to the view for final serialization.  The
        key names in this dict are not finalized at this stage, but all of the data is present.

        Each column is consulted for its value (computed based on its
        :py:attr:`~datatableview.columns.Column.sources` applied against the given ``obj`` instance)
        and then sent to the column's :py:attr:`~datatableview.columns.Column.processor` function.
        """

        preloaded_kwargs = self.preload_record_data(obj)
        data = {
            'pk': self.get_object_pk(obj),
            '_extra_data': self.get_extra_record_data(obj),
        }
        for i, (name, column) in enumerate(self.columns.items()):
            kwargs = dict(column.get_processor_kwargs(**preloaded_kwargs), **{
                'datatable': self,
                'view': self.view,
                'field_name': column.name,
            })
            value = self.get_column_value(obj, column, **kwargs)
            processor = self.get_processor_method(column, i)
            if processor:
                value = processor(obj, default_value=value[0], rich_value=value[1], **kwargs)

            # A 2-tuple at this stage has presumably served its purpose in the processor callback,
            # so we convert it to its "rich" value for display purposes.
            if isinstance(value, (tuple, list)):
                value = value[1]

            if six.PY2 and isinstance(value, str):  # not unicode
                value = value.decode('utf-8')
            if value is not None:
                value = six.text_type(value)
            data[str(i)] = value
        return data

    def get_column_value(self, obj, column, **kwargs):
        """ Returns whatever the column derived as the source value. """
        return column.value(obj, **kwargs)

    def get_processor_method(self, column, i):
        """
        Using a slightly mangled version of the column's name (explained below) each column's value
        is derived.

        Each field can generate customized data by defining a method on the view called either
        "get_column_FIELD_NAME_data" or "get_column_INDEX_data".

        If the FIELD_NAME approach is used, the name is the raw field name (e.g., "street_name") or
        else the friendly representation defined in a 2-tuple such as
        ("Street name", "subdivision__home__street_name"), where the name has non-alphanumeric
        characters stripped to single underscores.  For example, the friendly name
        "Region: Subdivision Type" would convert to "Region_Subdivision_Type", requiring the method
        name "get_column_Region_Subdivision_Type_data".

        Alternatively, if the INDEX approach is used, a method will be fetched called
        "get_column_0_data", or otherwise using the 0-based index of the column's position as
        defined in the view's ``datatable_options['columns']`` setting.

        Finally, if a third element is defined in the tuple, it will be treated as the function or
        name of a member attribute which will be used directly.

        """

        callback = column.processor
        if callback:
            if callable(callback):
                return callback
            if self.forward_callback_target:
                f = getattr(self.forward_callback_target, callback, None)
            else:
                f = None
            if f:
                return f
            return getattr(self, callback)

        column_name = column.name
        if isinstance(self, LegacyDatatable):
            name = force_text(column.label, errors="ignore")
            if not name:
                name = column.sources[0]
            column_name = re.sub(r'[\W_]+', '_', name)

        if self.forward_callback_target:
            f = getattr(self.forward_callback_target, 'get_column_%s_data' % (column_name,), None)
            if f:
                return f

            f = getattr(self.forward_callback_target, 'get_column_%d_data' % (i,), None)
            if f:
                return f

        f = getattr(self, 'get_column_%s_data' % (column_name,), None)
        if f:
            return f

        f = getattr(self, 'get_column_%d_data' % (i,), None)
        if f:
            return f

        return None


    # Template rendering features
    def __str__(self):
        """ Renders ``structure_template`` with ``self`` as a context variable. """

        if not hasattr(self, 'config'):
            self.configure()

        context = {
            'url': self.url,
            'config': self.config,
            'datatable': self,
            'columns': self.columns.values(),
        }
        return render_to_string(self.config['structure_template'], context)

    def __iter__(self):
        """ Yields each column in order. """

        if not hasattr(self, 'config'):
            self.configure()

        for column in self.columns.values():
            yield column


class ValuesDatatable(Datatable):
    """
    Variant of the standard Datatable that terminates its queryset with ``.values()`` provides the
    results to any column :py:attr:`~datatableview.columns.Column.processor` callbacks for
    additional modification.

    Processor callbacks will no longer receive model instances, but instead the dict of selected
    values.
    """
    def get_valuesqueryset(self, queryset):
        # Figure out the full list of ORM path names
        self.value_queries = OrderedDict({'pk': 'pk'})
        for name, column in self.columns.items():
            self.value_queries.update(OrderedDict([
                (source, name) for source in column.sources
            ]))

        return queryset.values(*self.value_queries.keys())

    def populate_records(self):
        """
        Switches the original queryset to a ``ValuesQuerySet``, selecting values according to what
        each column has declared in its :py:attr:`~datatableview.columns.Column.sources` list.
        """

        self.object_list = self.get_valuesqueryset(self.object_list)
        super(ValuesDatatable, self).populate_records()

    def get_object_pk(self, obj):
        """
        Correctly reads the pk from the ValuesQuerySet entry, as a dict item instead of an
        attribute.
        """
        return obj['pk']

    def preload_record_data(self, obj):
        """
        Modifies the ``obj`` values dict to alias the selected values to the column name that asked
        for its selection.

        For example, a datatable that declares a column ``'blog'`` which has a related lookup source
        ``'blog__name'`` will ensure that the selected value exists in ``obj`` at both keys
        ``blog__name`` and ``blog`` (the former because that was how it was selected, the latter
        because that was the column name used to select it).

        :Example:

            ``{'pk': 1, 'blog__name': "My Blog"}``
            ``{'pk': 1: 'blog__name': "My Blog", 'blog': "My Blog"}``

        When a column declares multiple :py:attr:`~datatableview.columns.Column.sources`, the column
        name's entry in ``obj`` will be a list of each of those values.

        :Example:

            ``{'pk': 1, 'blog__name': "My Blog", 'blog__id': 5}``
            ``{'pk': 1: 'blog__name': "My Blog", 'blog__id': 5, 'blog': ["My Blog", 5]}``

        In every situation, the original selected values will always be retained in ``obj``.
        """
        data = {}

        for orm_path, column_name in self.value_queries.items():
            value = obj[orm_path]
            if column_name not in data:
                data[column_name] = value
            else:
                if not isinstance(data[column_name], (tuple, list)):
                    data[column_name] = [data[column_name]]
                data[column_name].append(value)
        obj.update(data)
        return super(ValuesDatatable, self).preload_record_data(obj)


class LegacyDatatable(Datatable):
    """
    Modifies the :py:meth:`.resolve_virtual_columns` hook to deal with legacy-style column
    declarations, rather than automatically raising them as errors like normal.

    :py:class:`~datatableview.views.legacy.LegacyDatatableView` automatically uses this
    as its :py:attr:`~datatableview.views.legacy.LegacyDatatableView.datatable_class`.
    """

    def resolve_virtual_columns(self, *names):
        """
        Assume that all ``names`` are legacy-style tuple declarations, and generate modern columns
        instances to match the behavior of the old syntax.
        """
        from .views.legacy import get_field_definition
        virtual_columns = {}
        for name in names:
            field = get_field_definition(name)
            column = TextColumn(sources=field.fields, label=field.pretty_name,
                                processor=field.callback)
            column.name = field.pretty_name if field.pretty_name else field.fields[0]
            virtual_columns[name] = column

        # Make sure it's in the same order as originally defined
        new_columns = OrderedDict()
        for name in self._meta.columns:  # Can't use self.config yet, hasn't been generated
            if self.columns.get(name):
                column = self.columns[name]
            else:
                column = virtual_columns[name]
            new_columns[column.name] = column
        self.columns = new_columns


class ValuesLegacyDatatable(LegacyDatatable, ValuesDatatable):
    """
    A :py:class:`LegacyDatatable` that also inherits from :py:class:`ValuesDatatable`
    """



# -*- encoding: utf-8 -*-
""" Backports of code left behind by new versions of Django. """

import django
from django.utils.html import escape
try:
    from django.utils.encoding import escape_uri_path as django_escape_uri_path
except ImportError:
    django_escape_uri_path = None

import six


# Django 1.7 removed StrAndUnicode, so it has been purged from this project as well.  To bridge the
# gap, we will rely on this utility directly, instead of trying to generate our own replacement
# StrAndUnicode class.
def python_2_unicode_compatible(klass):
    """
    A decorator that defines __unicode__ and __str__ methods under Python 2.
    Under Python 3 it does nothing.

    To support Python 2 and 3 with a single code base, define a __str__ method
    returning text and apply this decorator to the class.
    """
    if not six.PY3:
        klass.__unicode__ = klass.__str__
        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return klass


def escape_uri_path(path):
    if django_escape_uri_path:
        return django_escape_uri_path(path)
    return escape(path)



import inspect
import hashlib
import logging

from django.core.cache import caches
from django.conf import settings

log = logging.getLogger(__name__)


class cache_types(object):
    NONE = None
    DEFAULT = 'default'
    SIMPLE = 'simple'  # Stores queryset objects directly in cache
    PK_LIST = 'pk_list'  # Stores queryset pks in cache for later expansion back to queryset


try:
    CACHE_BACKEND = settings.DATATABLEVIEW_CACHE_BACKEND
except AttributeError:
    CACHE_BACKEND = 'default'

try:
    CACHE_PREFIX = settings.DATATABLEVIEW_CACHE_PREFIX
except AttributeError:
    CACHE_PREFIX = 'datatableview_'

try:
    DEFAULT_CACHE_TYPE = settings.DATATABLEVIEW_DEFAULT_CACHE_TYPE
except AttributeError:
    DEFAULT_CACHE_TYPE = cache_types.SIMPLE

try:
    CACHE_KEY_HASH = settings.DATATABLEVIEW_CACHE_KEY_HASH
except AttributeError:
    CACHE_KEY_HASH = True

try:
    CACHE_KEY_HASH_LENGTH = settings.DATATABLEVIEW_CACHE_KEY_HASH_LENGTH
except AttributeError:
    CACHE_KEY_HASH_LENGTH = None

cache = caches[CACHE_BACKEND]


hash_slice = None
if CACHE_KEY_HASH:
    hash_slice = slice(None, CACHE_KEY_HASH_LENGTH)

def _hash_key_component(s):
    return hashlib.sha1(s.encode('utf-8')).hexdigest()[hash_slice]


def get_cache_key(datatable_class, view=None, user=None, **kwargs):
    """
    Returns a cache key unique to the current table, and (if available) the request user.

    The ``view`` argument should be the class reference itself, since it is easily obtainable
    in contexts where the instance is not available.
    """

    datatable_name = datatable_class.__name__
    if datatable_name.endswith('_Synthesized'):
        datatable_name = datatable_name[:-12]
    datatable_id = '%s.%s' % (datatable_class.__module__, datatable_name)
    if CACHE_KEY_HASH:
        datatable_id = _hash_key_component(datatable_id)

    cache_key = 'datatable_%s' % (datatable_id,)

    if view:
        if not inspect.isclass(view):
            # Reduce view to its class
            view = view.__class__

        view_id = '%s.%s' % (view.__module__, view.__name__)
        if CACHE_KEY_HASH:
            view_id = _hash_key_component(view_id)
        cache_key += '__view_%s' % (view_id,)

    if user and user.is_authenticated():
        cache_key += '__user_%s' % (user.pk,)

    # All other kwargs are used directly to create a hashed suffix
    # Order the kwargs by key name, then convert them to their repr() values.
    items = sorted(kwargs.items(), key=lambda item: item[0])
    values = []
    for k, v in items:
        values.append('%r:%r' % (k, v))

    if values:
        kwargs_id = '__'.join(values)
        kwargs_id = _hash_key_component(kwargs_id)
        cache_key += '__kwargs_%s' % (kwargs_id,)

    log.debug("Cache key derived for %r: %r (from kwargs %r)", datatable_class, cache_key, values)

    return cache_key


def get_cached_data(datatable, **kwargs):
    """ Returns the cached object list under the appropriate key, or None if not set. """
    cache_key = '%s%s' % (CACHE_PREFIX, datatable.get_cache_key(**kwargs))
    data = cache.get(cache_key)
    log.debug("Reading data from cache at %r: %r", cache_key, data)
    return data


def cache_data(datatable, data, **kwargs):
    """ Stores the object list in the cache under the appropriate key. """
    cache_key = '%s%s' % (CACHE_PREFIX, datatable.get_cache_key(**kwargs))
    log.debug("Setting data to cache at %r: %r", cache_key, data)
    cache.set(cache_key, data)



# -*- encoding: utf-8 -*-

import json

from django.urls import reverse

import six

from .testcase import DatatableViewTestCase
from .example_project.example_project.example_app import views
from .example_project.example_project.example_app import models


class FakeRequest(object):
    def __init__(self, url, method='GET'):
        self.path = url
        self.method = method
        setattr(self, method, {})


class ViewsTests(DatatableViewTestCase):
    urls = 'datatableview.tests.example_project.example_project.example_app.urls'

    fixtures = ['initial_data.json']

    def get_json_response(self, url):
        response = self.client.get(url, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        content = response.content
        if six.PY3:
            content = content.decode()
        return json.loads(content)

    def test_zero_configuration_datatable_view(self):
        """ Verifies that no column definitions means all local fields are used. """
        view = views.ZeroConfigurationDatatableView
        url = reverse('zero-configuration')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertEqual(
            len(list(response.context['datatable'])),
            len(models.Entry._meta.local_fields)
        )

    def test_specific_columns_datatable_view(self):
        """ Verifies that "columns" list matches context object length. """
        view = views.SpecificColumnsDatatableView()
        url = reverse('specific-columns')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertEqual(
            len(list(response.context['datatable'])),
            len(view.get_datatable().columns)
        )

    def test_pretty_names_datatable_view(self):
        """ Verifies that a pretty name definition is used instead of the verbose name. """
        view = views.PrettyNamesDatatableView()
        url = reverse('pretty-names')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertEqual(
            len(list(response.context['datatable'])),
            len(view.get_datatable().columns)
        )
        self.assertEqual(response.context['datatable'].columns['pub_date'].label, "Publication date")

    # def test_x_editable_columns_datatable_view(self):
    #     view = views.XEditableColumnsDatatableView
    #     url = reverse('x-editable-columns')
    #     response = self.client.get(url)

    def test_customized_template_datatable_view(self):
        """
        Verify that the custom structure template is getting rendered instead of the default one.
        """
        view = views.CustomizedTemplateDatatableView()
        url = reverse('customized-template')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertContains(response, """<table class="table table-striped table-bordered datatable" """)

    def test_bootstrap_template_datatable_view(self):
        """
        Verify that the custom structure template is getting rendered instead of the default one.
        """
        view = views.BootstrapTemplateDatatableView()
        url = reverse('bootstrap-template')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertContains(response, """<table class="display datatable table table-striped table-bordered" """)

    def test_multiple_tables_datatable_view(self):
        view = views.MultipleTablesDatatableView()
        url = reverse('multiple-tables')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertIn('demo1_datatable', response.context)
        self.assertIn('demo2_datatable', response.context)
        self.assertIn('demo3_datatable', response.context)

        demo1_obj = self.get_json_response(str(url) + "?datatable=demo1")
        self.assertEqual(len(demo1_obj['data'][0]), 2 + 2)  # 2 built-in DT items
        demo2_obj = self.get_json_response(str(url) + "?datatable=demo2")
        self.assertEqual(len(demo2_obj['data'][0]), 1 + 2)  # 2 built-in DT items
        demo3_obj = self.get_json_response(str(url) + "?datatable=demo3")
        self.assertEqual(len(demo3_obj['data'][0]), 3 + 2)  # 2 built-in DT items

    def test_embedded_table_datatable_view(self):
        view = views.SatelliteDatatableView()
        url = reverse('embedded-table')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        self.assertEqual(
            len(list(response.context['datatable'])),
            len(view.get_datatable().columns)
        )


    # Straightforward views that call on procedural logic not worth testing.  We would effectively
    # be proving that Python strings concatenate, etc.
    # Instead of proving details of the callbacks we've written, we'll just ask for the views, to
    # make certain that they don't generate errors.

    def test_column_backed_by_method_datatable_view(self):
        view = views.ColumnBackedByMethodDatatableView
        url = reverse('column-backed-by-method')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)

    def test_compound_columns_datatable_view(self):
        view = views.CompoundColumnsDatatableView
        url = reverse('compound-columns')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)

    def test_many_to_many_fields_datatable_view(self):
        view = views.ManyToManyFieldsDatatableView
        url = reverse('many-to-many-fields')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)

    def test_default_callback_names_datatable_view(self):
        view = views.DefaultCallbackNamesDatatableView
        url = reverse('default-callback-names')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)

    def test_helpers_reference_datatable_view(self):
        view = views.HelpersReferenceDatatableView
        url = reverse('helpers-reference')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)

    def test_satellite_datatable_view(self):
        view = views.SatelliteDatatableView
        url = reverse('satellite')
        view.request = FakeRequest(url)
        response = self.client.get(url)
        obj = self.get_json_response(url)



# -*- encoding: utf-8 -*-

import json
import logging

from django.views.generic import ListView, TemplateView
from django.views.generic.list import MultipleObjectMixin
from django.http import HttpResponse
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

from ..datatables import Datatable, DatatableOptions
from ..compat import escape_uri_path

log = logging.getLogger(__name__)


class DatatableJSONResponseMixin(object):
    def dispatch(self, request, *args, **kwargs):
        if request.is_ajax() or getattr(request, request.method).get('ajax') == 'true':
            datatable = self.get_datatable(request)
            datatable.configure()
            if request.method == datatable.config['request_method']:
                return self.get_ajax(request, *args, **kwargs)
        return super(DatatableJSONResponseMixin, self).dispatch(request, *args, **kwargs)

    # Response generation
    def get_json_response_object(self, datatable):
        """
        Returns the JSON-compatible dictionary that will be serialized for an AJAX response.

        The value names are in the form "s~" for strings, "i~" for integers, and "a~" for arrays,
        if you're unfamiliar with the old C-style jargon used in dataTables.js.  "aa~" means
        "array of arrays".  In some instances, the author uses "ao~" for "array of objects", an
        object being a javascript dictionary.
        """

        # Ensure the object list is calculated.
        # Calling get_records() will do this implicitly, but we want simultaneous access to the
        # 'total_initial_record_count', and 'unpaged_record_count' values.
        datatable.populate_records()

        draw = getattr(self.request, self.request.method).get('draw', None)
        if draw is not None:
            draw = escape_uri_path(draw)
        response_data = {
            'draw': draw,
            'recordsFiltered': datatable.unpaged_record_count,
            'recordsTotal': datatable.total_initial_record_count,
            'data': [dict(record, **{
                'DT_RowId': record.pop('pk'),
                'DT_RowData': record.pop('_extra_data'),
            }) for record in datatable.get_records()],
        }
        return response_data

    def serialize_to_json(self, response_data):
        """ Returns the JSON string for the compiled data object. """

        indent = None
        if settings.DEBUG:
            indent = 4

        # Serialize to JSON with Django's encoder: Adds date/time, decimal,
        # and UUID support.
        return json.dumps(response_data, indent=indent, cls=DjangoJSONEncoder)


class DatatableMixin(DatatableJSONResponseMixin, MultipleObjectMixin):
    """
    Adds a JSON AJAX response mechanism that can be used by the datatables.js code to load
    server-side records.
    """

    datatable_class = None
    context_datatable_name = 'datatable'

    # AJAX response handler
    def get_ajax(self, request, *args, **kwargs):
        """ Called when accessed via AJAX on the request method specified by the Datatable. """
        
        response_data = self.get_json_response_object(self._datatable)
        response = HttpResponse(self.serialize_to_json(response_data),
                                content_type="application/json")
        return response

    # Configuration getters
    def get_datatable(self, request, **kwargs):
        """ Gathers and returns the final :py:class:`Datatable` instance for processing. """
        if hasattr(self, '_datatable'):
            return self._datatable
        
        datatable_class = self.get_datatable_class()
        if datatable_class is None:
            class AutoMeta:
                model = self.model or self.get_queryset().model
            opts = AutoMeta()
            datatable_class = Datatable
        else:
            opts = datatable_class.options_class(datatable_class._meta)
        kwargs = self.get_datatable_kwargs(request, **kwargs)
        for meta_opt in opts.__dict__:
            if meta_opt in kwargs:
                setattr(opts, meta_opt, kwargs.pop(meta_opt))

        datatable_class = type('%s_Synthesized' % (datatable_class.__name__,), (datatable_class,), {
            '__module__': datatable_class.__module__,
            'Meta': opts,
        })
        
        self._datatable = datatable_class(**kwargs)
        return self._datatable

    def get_datatable_class(self):
        return self.datatable_class

    def get_datatable_kwargs(self, request, **kwargs):
        kwargs['user'] = request.user
        queryset = self.get_queryset(**kwargs)
        kwargs.update({
            'object_list': queryset,
            'view': self,
            'model': self.model or queryset.model,
        })
        
        # This is, i.e., request, provided by default, but if the view is instantiated outside of the request cycle
        # (such as for the purposes of embedding that view's datatable elsewhere), the request may
        # not be required, so the user may not have a compelling reason to go through the trouble of
        # putting it on self.
        if hasattr(self, 'request'):
            kwargs['url'] = self.request.path
            kwargs['query_config'] = getattr(self.request, self.request.method)
            kwargs['user'] = self.request.user
            
        
        else:
            kwargs['query_config'] = {}

        settings = ('columns', 'exclude', 'ordering', 'start_offset', 'page_length', 'search',
                    'search_fields', 'unsortable_columns', 'hidden_columns', 'footer',
                    'structure_template', 'result_counter_id')

        for k in settings:
            v = getattr(self, k, None)
            if v is not None:  # MultipleObjectMixin or others might have default attr as None
                kwargs[k] = v
        
        return kwargs


    # Runtime per-object hook
    def preload_record_data(self, obj):
        return {}

    # Extra getters
    def get_datatable_context_name(self):
        return self.context_datatable_name

    def get_context_data(self, **kwargs):
        context = super(DatatableMixin, self).get_context_data(**kwargs)

        context[self.get_datatable_context_name()] = self.get_datatable()
        return context


class DatatableView(DatatableMixin, ListView):
    """ Implements :py:class:`DatatableMixin` and the standard Django ``ListView``. """


class MultipleDatatableMixin(DatatableJSONResponseMixin):
    """
    Allow multiple Datatable classes to be given as a dictionary of context names to classes.

    Methods will be dynamically inspected to supply the classes with a queryset and their
    initialization kwargs, in the form of ``get_FOO_datatable_queryset(**kwargs)`` or
    ``get_FOO_datatable_kwargs(**kwargs)`` respectively.

    In the case of the kwargs getter, the default generated kwargs can be retrieved via a call to
    ``get_default_datatable_kwargs(**kwargs)``, where ``**kwargs`` is a reference to the kwargs that
    came into the ``get_FOO_datatable_kwargs(**kwargs)`` method.
    """

    datatable_classes = None  # Dict of context names to class names

    # AJAX response handler
    def get_ajax(self, request, *args, **kwargs):
        """ Called in place of normal ``get()`` when accessed via AJAX. """

        response_data = self.get_json_response_object(self._datatable)
        response = HttpResponse(self.serialize_to_json(response_data),
                                content_type="application/json")

        return response

    # Configuration getters
    def get_datatable(self):
        if hasattr(self, '_datatable'):
            return self._datatable
        self._datatable = self.get_active_ajax_datatable()
        return self._datatable

    def get_active_ajax_datatable(self):
        """ Returns a single datatable according to the hint GET variable from an AJAX request. """
        data = getattr(self.request, self.request.method)
        datatables_dict = self.get_datatables(only=data['datatable'])
        return list(datatables_dict.values())[0]

    def get_datatables(self, only=None):
        """ Returns a dict of the datatables served by this view. """
        if not hasattr(self, '_datatables'):
            self._datatables = {}
            datatable_classes = self.get_datatable_classes()
            for name, datatable_class in datatable_classes.items():
                if only and name != only:
                    continue
                queryset_getter_name = 'get_%s_datatable_queryset' % (name,)
                queryset_getter = getattr(self, queryset_getter_name, None)
                if queryset_getter is None:
                    raise ValueError("%r must declare a method %r." % (self.__class__.__name__,
                                                                       queryset_getter_name))

                queryset = queryset_getter()
                if datatable_class is None:
                    class AutoMeta:
                        model = queryset.model
                    opts = AutoMeta()
                    datatable_class = Datatable
                else:
                    opts = datatable_class.options_class(datatable_class._meta)

                kwargs = self.get_default_datatable_kwargs(object_list=queryset)
                kwargs_getter_name = 'get_%s_datatable_kwargs' % (name,)
                kwargs_getter = getattr(self, kwargs_getter_name, None)
                if kwargs_getter:
                    kwargs = kwargs_getter(**kwargs)
                if 'url' in kwargs:
                    kwargs['url'] = kwargs['url'] + "?datatable=%s" % (name,)

                for meta_opt in opts.__dict__:
                    if meta_opt in kwargs:
                        setattr(opts, meta_opt, kwargs.pop(meta_opt))

                datatable_class = type('%s_Synthesized' % (datatable_class.__name__,), (datatable_class,), {
                    '__module__': datatable_class.__module__,
                    'Meta': opts,
                })

                self._datatables[name] = datatable_class(**kwargs)
        return self._datatables

    def get_datatable_classes(self):
        """ Return a shallow copy of the view's ``datatable_classes`` dict. """
        if self.datatable_classes is None:
            return {}
        return dict(self.datatable_classes)

    def get_default_datatable_kwargs(self, **kwargs):
        """
        Builds the default set of kwargs for initializing a Datatable class.  Note that by default
        the MultipleDatatableMixin does not support any configuration via the view's class
        attributes, and instead relies completely on the Datatable class itself to declare its
        configuration details.
        """

        kwargs['view'] = self

        # This is provided by default, but if the view is instantiated outside of the request cycle
        # (such as for the purposes of embedding that view's datatable elsewhere), the request may
        # not be required, so the user may not have a compelling reason to go through the trouble of
        # putting it on self.
        if hasattr(self, 'request'):
            kwargs['url'] = self.request.path
            kwargs['query_config'] = getattr(self.request, self.request.method)
        else:
            kwargs['query_config'] = {}

        return kwargs

    # Extra getters
    def get_context_data(self, **kwargs):
        context = super(MultipleDatatableMixin, self).get_context_data(**kwargs)

        for name, datatable in self.get_datatables().items():
            context_name = '%s_datatable' % (name,)
            context[context_name] = datatable

        return context


class MultipleDatatableView(MultipleDatatableMixin, TemplateView):
    pass



# -*- encoding: utf-8 -*-
from inspect import isgenerator

from .testcase import DatatableViewTestCase
from .test_app import models
from ..exceptions import ColumnError
from ..datatables import Datatable, ValuesDatatable
from ..views import DatatableJSONResponseMixin, DatatableView
from .. import columns

class DatatableTests(DatatableViewTestCase):
    def test_normalize_config(self):
        dt = Datatable([], '/')
        dt.configure()
        self.assertEqual(dt.config['hidden_columns'], [])
        self.assertEqual(dt.config['search_fields'], [])
        self.assertEqual(dt.config['unsortable_columns'], [])
        self.assertEqual(dt.config['search'], set())
        self.assertEqual(dt.config['start_offset'], 0)
        self.assertEqual(dt.config['page_length'], 25)
        self.assertEqual(dt.config['ordering'], None)

    def test_column_names_list_raises_unknown_columns(self):
        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['fake']

        dt = DT([], '/')
        with self.assertRaises(ColumnError) as cm:
            dt.configure()
        self.assertEqual(str(cm.exception), "Unknown column name(s): ('fake',)")

    def test_column_names_list_finds_local_fields(self):
        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['name']

        class NoError(BaseException):
            pass

        with self.assertRaises(NoError):
            dt = DT([], '/')
            raise NoError()

    def test_column_names_list_raises_related_columns(self):
        # This was the old way of including related data, but this is no longer supported
        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['related__name']

        dt = DT([], '/')
        with self.assertRaises(ColumnError) as cm:
            dt.configure()
        self.assertEqual(str(cm.exception), "Unknown column name(s): ('related__name',)")

    def test_column_names_list_finds_related_fields(self):
        class DT(Datatable):
            related = columns.TextColumn("Related", ['related__name'])
            class Meta:
                model = models.ExampleModel
                columns = ['name', 'related']

        class NoError(BaseException):
            pass

        with self.assertRaises(NoError):
            dt = DT([], '/')
            raise NoError()

    def test_get_ordering_splits(self):
        # Verify empty has blank db-backed list and virtual list
        dt = Datatable([], '/')
        dt.configure()
        self.assertEqual(dt.get_ordering_splits(), ([], []))

        class DT(Datatable):
            fake = columns.TextColumn("Fake", sources=['get_absolute_url'])

            class Meta:
                model = models.ExampleModel
                columns = ['name', 'fake']

        # Verify a fake field name ends up separated from the db-backed field
        dt = DT([], '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # iSortingCols': '1',
        dt.configure()
        self.assertEqual(dt.get_ordering_splits(), (['name'], []))

        # Verify ['name', 'fake'] ordering sends 'name' to db sort list, but keeps 'fake' in manual
        # sort list.
        dt = DT([], '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc', 'order[1][column]': '1', 'order[1][dir]': 'asc'}) # 'iSortingCols': '2',
        dt.configure()
        self.assertEqual(dt.get_ordering_splits(), (['name'], ['fake']))

        # Verify a fake field name as the sort column correctly finds no db sort fields
        dt = DT([], '/', query_config={'order[0][column]': '1', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.configure()
        self.assertEqual(dt.get_ordering_splits(), ([], ['fake']))

        # Verify ['fake', 'name'] ordering sends both fields to manual sort list
        dt = DT([], '/', query_config={'order[0][column]': '1', 'order[0][dir]': 'asc', 'order[1][column]': '0', 'order[1][dir]': 'asc'}) # 'iSortingCols': '2',
        dt.configure()
        self.assertEqual(dt.get_ordering_splits(), ([], ['fake', 'name']))

    def test_get_records_populates_cache(self):
        models.ExampleModel.objects.create(name="test name")
        queryset = models.ExampleModel.objects.all()

        dt = Datatable(queryset, '/')
        dt.get_records()
        self.assertIsNotNone(dt._records)
        records = dt._records

        # _records doesn't change when run again
        dt.get_records()
        self.assertEqual(dt._records, records)

    def test_populate_records_searches(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1", value=False)
        obj2 = models.ExampleModel.objects.create(name="test name 2", value=True)
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['name', 'value']
        dt = DT(queryset, '/')

        # Sanity check for correct initial queryset
        dt.populate_records()
        self.assertIsNotNone(dt._records)
        self.assertEqual(list(dt._records), list(queryset))

        # Verify a search eliminates items from _records
        dt = DT(queryset, '/', query_config={'search[value]': 'test name 1'})
        dt.populate_records()
        self.assertIsNotNone(dt._records)
        self.assertEqual(list(dt._records), [obj1])

    def test_populate_records_sorts(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        obj2 = models.ExampleModel.objects.create(name="test name 2")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['name']
        dt = DT(queryset, '/')

        # Sanity check for correct initial queryset
        dt.populate_records()
        self.assertIsNotNone(dt._records)
        self.assertEqual(list(dt._records), list(queryset))

        # Verify a sort changes the ordering of the records list
        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # # 'iSortingCols': '1',
        dt.populate_records()
        self.assertIsNotNone(dt._records)
        self.assertEqual(list(dt._records), [obj2, obj1])

    def test_populate_records_avoids_column_callbacks(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()
        class DT(Datatable):
            def preload_record_data(self, obj):
                raise Exception("Don't run this")
        dt = DT(queryset, '/')
        try:
            dt.populate_records()
        except Exception as e:
            if str(e) == "Don't run this":
                raise AssertionError("Per-row callbacks being executed!")
            raise

    def test_preload_record_data_calls_view(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()
        class Dummy(object):
            def preload_record_data(self, obj):
                raise Exception("We did it")
        dt = Datatable(queryset, '/', callback_target=Dummy())
        with self.assertRaises(Exception) as cm:
            dt.get_records()
        self.assertEqual(str(cm.exception), "We did it")

    def test_sort_defaults_to_meta_ordering(self):
        # Defined so that 'pk' order != 'name' order
        obj1 = models.ExampleModel.objects.create(name="b")
        obj2 = models.ExampleModel.objects.create(name="a")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            name = columns.TextColumn("Name", sources=['name'])
            class Meta:
                model = models.ExampleModel
                columns = ['name']
                ordering = ['name']

        dt = DT(queryset, '/')
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['name'], []))
        self.assertEqual(list(dt._records), [obj2, obj1])


        # this is to keep DatatableView class from overriding the Meta ordering in Datatable
        class DTV(DatatableView):
            datatable_class = DT
            model = models.ExampleModel

        dtv = DTV().get_datatable(url='/')
        self.assertIn('<th data-name="name" data-config-sortable="true" data-config-sorting="0,0,asc" data-config-visible="true">Name</th>', dtv.__str__())


        class DT(Datatable):
            name = columns.TextColumn("Name", sources=['name'])
            class Meta:
                model = models.ExampleModel
                columns = ['name']
                ordering = ['-name']

        dt = DT(queryset, '/')
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['-name'], []))
        self.assertEqual(list(dt._records), [obj1, obj2])

    def test_sort_prioritizes_db_source(self):
        # Defined so that 'pk' order != 'name' order
        obj1 = models.ExampleModel.objects.create(name="test name 2")
        obj2 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            name = columns.TextColumn("Name", sources=['name'])
            class Meta:
                model = models.ExampleModel
                columns = ['name']
                ordering = ['pk']

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['name'], []))
        self.assertEqual(list(dt._records), [obj2, obj1])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['-name'], []))
        self.assertEqual(list(dt._records), [obj1, obj2])

    def test_sort_uses_all_sources(self):
        from datetime import timedelta
        obj1 = models.ExampleModel.objects.create(name="a")
        obj2 = models.ExampleModel.objects.create(name="a")
        obj3 = models.ExampleModel.objects.create(name="b")
        obj1.date_created = obj1.date_created + timedelta(days=3)
        obj2.date_created = obj2.date_created + timedelta(days=1)
        obj3.date_created = obj3.date_created + timedelta(days=2)
        obj1.save()
        obj2.save()
        obj3.save()

        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            my_column = columns.TextColumn("Data", sources=['name', 'date_created', 'pk'])
            class Meta:
                model = models.ExampleModel
                columns = ['my_column']

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['my_column'], []))
        self.assertEqual(list(dt._records), [obj2, obj1, obj3])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['-my_column'], []))
        self.assertEqual(list(dt._records), [obj3, obj1, obj2])


        # Swap the order of 'date_created' and 'name' fields in the sources, which will alter the
        # sort results.
        class DT(Datatable):
            my_column = columns.TextColumn("Data", sources=['date_created', 'name', 'pk'])
            class Meta:
                model = models.ExampleModel
                columns = ['my_column']

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['my_column'], []))
        self.assertEqual(list(dt._records), [obj2, obj3, obj1])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['-my_column'], []))
        self.assertEqual(list(dt._records), [obj1, obj3, obj2])

    def test_sort_ignores_virtual_sources_when_mixed(self):
        from datetime import timedelta
        obj1 = models.ExampleModel.objects.create(name="a")
        obj2 = models.ExampleModel.objects.create(name="b")
        obj3 = models.ExampleModel.objects.create(name="a")

        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            my_column = columns.TextColumn("Data", sources=['name', 'get_absolute_url'])
            class Meta:
                model = models.ExampleModel
                columns = ['my_column']

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['my_column'], []))
        self.assertEqual(list(dt._records), [obj1, obj3, obj2])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), (['-my_column'], []))
        self.assertEqual(list(dt._records), [obj2, obj1, obj3])  # pk is natural ordering 1,3 here

        # Swap the sources order, but we expect the same result
        class DT(Datatable):
            my_column = columns.TextColumn("Data", sources=['get_absolute_url', 'name'], processor='get_data')
            class Meta:
                model = models.ExampleModel
                columns = ['my_column']

            def get_data(self, obj, **kwargs):
                # Return data that would make the sort order wrong if it were consulted for sorting
                return obj.pk  # tracks with get_absolute_url

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(list(dt._records), [obj1, obj3, obj2])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(list(dt._records), [obj2, obj1, obj3])  # pk is natural ordering 1,3 here

    def test_sort_uses_virtual_sources_when_no_db_sources_available(self):
        from datetime import timedelta
        obj1 = models.ExampleModel.objects.create(name="a")
        obj2 = models.ExampleModel.objects.create(name="b")
        obj3 = models.ExampleModel.objects.create(name="c")

        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            pk = columns.TextColumn("Data", sources=['get_negative_pk'])
            class Meta:
                model = models.ExampleModel
                columns = ['pk']

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'asc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), ([], ['pk']))
        self.assertEqual(list(dt._records), [obj3, obj2, obj1])

        dt = DT(queryset, '/', query_config={'order[0][column]': '0', 'order[0][dir]': 'desc'}) # 'iSortingCols': '1',
        dt.populate_records()
        self.assertEqual(dt.get_ordering_splits(), ([], ['-pk']))
        self.assertEqual(list(dt._records), [obj1, obj2, obj3])

    def test_get_object_pk(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()
        dt = Datatable(queryset, '/')
        self.assertEqual(dt.get_object_pk(obj1), obj1.pk)

    def test_get_extra_record_data_passes_through_to_object_serialization(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            def get_extra_record_data(self, obj):
                return {'custom': 'data'}

        dt = DT([], '/')
        data = dt.get_record_data(obj1)
        self.assertIn('_extra_data', data)
        self.assertIn('custom', data['_extra_data'])
        self.assertEqual(data['_extra_data']['custom'], 'data')

    def test_get_extra_record_data_passes_through_to_json_response(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            def get_extra_record_data(self, obj):
                return {'custom': 'data'}

        class FakeRequest(object):
            method = 'GET'
            GET = {'sEcho': 0}

        dt = DT(queryset, '/')
        view = DatatableJSONResponseMixin()
        view.request = FakeRequest()
        data = view.get_json_response_object(dt)
        self.assertIn('data', data)
        self.assertIn('DT_RowData', data['data'][0])
        self.assertEqual(data['data'][0]['DT_RowData'], {'custom': 'data'})

    def test_get_column_value_forwards_to_column_class(self):
        class CustomColumn1(columns.Column):
            def value(self, obj, **kwargs):
                return "first"

        class CustomColumn2(columns.Column):
            def value(self, obj, **kwargs):
                return "second"

        class DT(Datatable):
            fake1 = CustomColumn1("Fake1", sources=['get_absolute_url'])
            fake2 = CustomColumn2("Fake2", sources=['get_absolute_url'])

            class Meta:
                model = models.ExampleModel
                columns = ['name', 'fake1', 'fake2']

        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()
        dt = DT(queryset, '/')
        data = dt.get_record_data(obj1)
        self.assertIn('1', data)
        self.assertIn(data['1'], 'first')
        self.assertIn('2', data)
        self.assertIn(data['2'], 'second')

    def test_get_processor_method(self):
        class Dummy(object):
            def fake_callback(self):
                pass

        view = Dummy()

        # Test no callback given
        dt = Datatable([], '/')
        f = dt.get_processor_method(columns.Column("Fake", sources=['fake']), i=0)
        self.assertEqual(f, None)

        class DT(Datatable):
            def fake_callback(self):
                pass

        column = columns.Column("Fake", sources=['fake'], processor='fake_callback')

        # Test callback found on self
        dt = DT([], '/')
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, dt.fake_callback)

        # Test callback found on callback_target
        dt = Datatable([], '/', callback_target=view)
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, view.fake_callback)

    def test_get_processor_method_returns_direct_callable(self):
        def fake_callback():
            pass

        column = columns.Column("Fake", sources=[], processor=fake_callback)

        # Test no callback given
        dt = Datatable([], '/')
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, fake_callback)

    def test_get_processor_method_finds_implied_callback(self):
        class DummyNamed(object):
            def get_column_fake_data(self):
                pass

        class DummyIndexed(object):
            def get_column_0_data(self):
                pass

        class DummyBoth(object):
            def get_column_fake_data(self):
                pass

            def get_column_0_data(self):
                pass

        column = columns.Column("Fake", sources=[])
        column.name = 'fake'

        # Test implied named callback found first
        view = DummyNamed()
        dt = Datatable([], '/', callback_target=view)
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, view.get_column_fake_data)

        # Test implied named callback found first
        view = DummyIndexed()
        dt = Datatable([], '/', callback_target=view)
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, view.get_column_0_data)

        # Test implied named callback found first
        view = DummyBoth()
        dt = Datatable([], '/', callback_target=view)
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, view.get_column_fake_data)

        class DTNamed(Datatable):
            def get_column_fake_data(self):
                pass

        class DTIndexed(Datatable):
            def get_column_0_data(self):
                pass

        class DTBoth(Datatable):
            def get_column_fake_data(self):
                pass

            def get_column_0_data(self):
                pass


        # Test implied named callback found first
        dt = DTNamed([], '/')
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, dt.get_column_fake_data)

        # Test implied named callback found first
        dt = DTIndexed([], '/')
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, dt.get_column_0_data)

        # Test implied named callback found first
        dt = DTBoth([], '/')
        f = dt.get_processor_method(column, i=0)
        self.assertEqual(f, dt.get_column_fake_data)

    def test_iter_datatable_yields_columns(self):
        class CustomColumn1(columns.Column):
            pass

        class CustomColumn2(columns.Column):
            pass

        class DT(Datatable):
            fake1 = CustomColumn1("Fake1", sources=['get_absolute_url'])
            fake2 = CustomColumn2("Fake2", sources=['get_absolute_url'])

            class Meta:
                model = models.ExampleModel
                columns = ['name', 'fake1', 'fake2']

        dt = DT([], '/')
        self.assertEqual(isgenerator(dt.__iter__()), True)
        self.assertEqual(list(dt), [dt.columns['name'], dt.columns['fake1'], dt.columns['fake2']])

    def test_search_term_basic(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        obj2 = models.ExampleModel.objects.create(name="test name 2")
        obj3 = models.ExampleModel.objects.create(name="test name 12")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['name']

        dt = DT(queryset, '/', query_config={'search[value]': 'test'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': 'name'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': '1'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': '2'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': '12'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj3])

        dt = DT(queryset, '/', query_config={'search[value]': '3'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [])

    def test_search_term_boolean(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1", value=True)
        obj2 = models.ExampleModel.objects.create(name="test name 2", value=True)
        obj3 = models.ExampleModel.objects.create(name="test name 12", value=False)
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            senior = columns.BooleanColumn('Senior:', 'value')
            class Meta:
                model = models.ExampleModel
                columns = ['name', 'senior']

        dt = DT(queryset, '/', query_config={'search[value]': 'True'})
        dt.populate_records()
        self.assertEquals(len(list(dt._records)), 2)

        dt = DT(queryset, '/', query_config={'search[value]': 'false'})
        dt.populate_records()
        self.assertEquals(len(list(dt._records)), 1)

        dt = DT(queryset, '/', query_config={'search[value]': 'SENIOR'})
        dt.populate_records()
        self.assertEquals(len(list(dt._records)), 2)

        dt = DT(queryset, '/', query_config={'search[value]': 'menior'})
        dt.populate_records()
        self.assertEquals(len(list(dt._records)), 0)

    def test_search_multiple_terms_use_AND(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        obj2 = models.ExampleModel.objects.create(name="test name 2")
        obj3 = models.ExampleModel.objects.create(name="test name 12")
        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            class Meta:
                model = models.ExampleModel
                columns = ['name']

        dt = DT(queryset, '/', query_config={'search[value]': 'test name'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': 'test 1'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': 'test 2'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2, obj3])

        dt = DT(queryset, '/', query_config={'search[value]': 'test 12'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj3])

        dt = DT(queryset, '/', query_config={'search[value]': 'test 3'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [])

    def test_search_term_queries_all_columns(self):
        r1 = models.RelatedModel.objects.create(name="test related 1 one")
        r2 = models.RelatedModel.objects.create(name="test related 2 two")
        obj1 = models.ExampleModel.objects.create(name="test name 1", related=r1)
        obj2 = models.ExampleModel.objects.create(name="test name 2", related=r2)

        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            related = columns.TextColumn("Related", ['related__name'])
            class Meta:
                model = models.ExampleModel
                columns = ['name', 'related']

        dt = DT(queryset, '/', query_config={'search[value]': 'test'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'test name'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'test 2'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'related 2'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'test one'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1])

        dt = DT(queryset, '/', query_config={'search[value]': '2 two'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'test three'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [])

    def test_search_term_queries_extra_fields(self):
        r1 = models.RelatedModel.objects.create(name="test related 1 one")
        r2 = models.RelatedModel.objects.create(name="test related 2 two")
        obj1 = models.ExampleModel.objects.create(name="test name 1", related=r1)
        obj2 = models.ExampleModel.objects.create(name="test name 2", related=r2)

        queryset = models.ExampleModel.objects.all()

        class DT(Datatable):
            related = columns.TextColumn("Related", ['related__name'])
            class Meta:
                model = models.ExampleModel
                columns = ['related']
                search_fields = ['name']

        dt = DT(queryset, '/', query_config={'search[value]': 'test'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj1, obj2])

        dt = DT(queryset, '/', query_config={'search[value]': 'test name 2'})
        dt.populate_records()
        self.assertEquals(list(dt._records), [obj2])


class ValuesDatatableTests(DatatableViewTestCase):
    def test_get_object_pk(self):
        obj1 = models.ExampleModel.objects.create(name="test name 1")
        queryset = models.ExampleModel.objects.all()
        dt = ValuesDatatable(queryset, '/')
        obj_data = queryset.values('pk')[0]
        self.assertEqual(dt.get_object_pk(obj_data), obj1.pk)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2016-10-05 00:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Author',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('email', models.EmailField(max_length=254)),
            ],
        ),
        migrations.CreateModel(
            name='Blog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('tagline', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('headline', models.CharField(max_length=255)),
                ('body_text', models.TextField()),
                ('pub_date', models.DateField()),
                ('mod_date', models.DateField()),
                ('n_comments', models.IntegerField()),
                ('n_pingbacks', models.IntegerField()),
                ('rating', models.IntegerField()),
                ('status', models.IntegerField(choices=[(0, b'Draft'), (1, b'Published')])),
                ('authors', models.ManyToManyField(to='example_app.Author')),
                ('blog', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='example_app.Blog')),
            ],
        ),
    ]



# -*- encoding: utf-8 -*-

import re
import operator
from datetime import datetime
try:
    from functools import reduce
except ImportError:
    pass

import django
from django.db import models
from django.db.models import Model, Manager, Q
from django.db.models.fields import FieldDoesNotExist
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_text
from django.utils.safestring import mark_safe
try:
    from django.forms.utils import flatatt
except ImportError:
    from django.forms.util import flatatt
from django.template.defaultfilters import slugify
try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError:
    from .compat import python_2_unicode_compatible

import six
import dateutil.parser

from .utils import resolve_orm_path, DEFAULT_EMPTY_VALUE, DEFAULT_MULTIPLE_SEPARATOR

# Registry of Column subclasses to their declared corresponding ModelFields.
# The registery is an ordered priority list, containing 2-tuples of a Column subclass and a list of
# classes that the column will service.
COLUMN_CLASSES = []

STRPTIME_PLACEHOLDERS = {
    'year': ('%y', '%Y'),
    'month': ('%m', '%b', '%B'),
    'day': ('%d',),# '%a', '%A'),  # day names are hard because they depend on other date info
    'hour': ('%H', '%I'),
    'minute': ('%M',),
    'second': ('%S',),
    'week_day': ('%w',),
}

def register_simple_modelfield(model_field):
    column_class = get_column_for_modelfield(model_field)
    COLUMN_CLASSES.insert(0, (column_class, [model_field]))

def get_column_for_modelfield(model_field):
    """ Return the built-in Column class for a model field class. """

    # If the field points to another model, we want to get the pk field of that other model and use
    # that as the real field.  It is possible that a ForeignKey points to a model with table
    # inheritance, however, so we need to traverse the internal OneToOneField as well, so this will
    # climb the 'pk' field chain until we have something real.
    while model_field.related_model:
        model_field = model_field.related_model._meta.pk
    for ColumnClass, modelfield_classes in COLUMN_CLASSES:
        if isinstance(model_field, tuple(modelfield_classes)):
            return ColumnClass

def get_attribute_value(obj, bit):
    try:
        value = getattr(obj, bit)
    except (AttributeError, ObjectDoesNotExist):
        value = None
    else:
        if callable(value) and not isinstance(value, Manager):
            if not hasattr(value, 'alters_data') or value.alters_data is not True:
                value = value()
    return value

class ColumnMetaclass(type):
    """ Column type for automatic registration of column types as ModelField handlers. """
    def __new__(cls, name, bases, attrs):
        new_class = super(ColumnMetaclass, cls).__new__(cls, name, bases, attrs)
        if new_class.model_field_class:
            COLUMN_CLASSES.insert(0, (new_class, [new_class.model_field_class]))
            if new_class.handles_field_classes:
                COLUMN_CLASSES.insert(0, (new_class, new_class.handles_field_classes))
        return new_class


# Corollary to django.forms.fields.Field
@python_2_unicode_compatible
class Column(six.with_metaclass(ColumnMetaclass)):
    """ Generic table column using CharField for rendering. """

    model_field_class = None
    handles_field_classes = []

    lookup_types = ()

    # Tracks each time a Field instance is created. Used to retain order.
    creation_counter = 0

    def __init__(self, label=None, sources=None, processor=None, source=None,
                 separator=DEFAULT_MULTIPLE_SEPARATOR, empty_value=DEFAULT_EMPTY_VALUE,
                 model_field_class=None, sortable=True, visible=True, localize=False,
                 allow_regex=False, allow_full_text_search=False):
        if model_field_class:
            self.model_field_class = model_field_class

        if source and sources:
            raise ValueError("Cannot provide 'source' and 'sources' at the same time.")

        if source:
            sources = source

        self.name = None  # Set outside, once the Datatable can put it there
        if label is not None:
            label = smart_text(label)
        self.sources = sources or []  # TODO: Process for real/virtual
        if not isinstance(self.sources, (tuple, list)):
            self.sources = [self.sources]
        self.separator = separator
        self.label = label
        self.empty_value = smart_text(empty_value)
        self.localize = localize
        self.sortable = sortable
        self.visible = visible
        self.processor = processor
        self.allow_regex = allow_regex
        self.allow_full_text_search = allow_full_text_search

        if not self.sources:
            self.sortable = False

        # To be filled in externally once the datatable has ordering figured out.
        self.sort_priority = None
        self.sort_direction = None
        self.index = None

        # Increase the creation counter, and save our local copy.
        self.creation_counter = Column.creation_counter
        Column.creation_counter += 1

    def __repr__(self):
        return '<%s.%s "%s">' % (self.__class__.__module__, self.__class__.__name__, self.label)

    def value(self, obj, **kwargs):
        """
        Calls :py:meth:`.get_initial_value` to obtain the value from ``obj`` that this column's
        :py:attr:`.sources` list describes.

        Any supplied ``kwargs`` are forwarded to :py:meth:`.get_initial_value`.

        Returns the 2-tuple of ``(plain_value, rich_value)`` for the inspection and serialization
        phases of serialization.
        """

        values = self.get_initial_value(obj, **kwargs)

        if not isinstance(values, (tuple, list)):
            values = (values, values)

        return values

    def get_initial_value(self, obj, **kwargs):
        """
        Builds a list of values provided by :py:attr:`.sources` looked up on the target ``obj``.
        Each source may provide a value as a 2-tuple of ``(plain_value, rich_value)``, where
        ``plain_value`` is the sortable raw value, and ``rich_value`` is possibly something else
        that can be coerced to a string for display purposes. The ``rich_value`` could also be a
        string with HTML in it.

        If no 2-tuple is given, then ``plain_value`` and ``rich_value`` are taken to be the same.

        Columns with multiple :py:attr:`.sources` will have their ``rich_value`` coerced to a string
        and joined with :py:attr:`.separator`, and this new concatenated string becomes the final
        ``rich_value`` for the whole column.

        If all :py:attr:`.sources` are ``None``, :py:attr:`.empty_value` will be used as the
        ``rich_value``.
        """

        values = []
        for source in self.sources:
            result = self.get_source_value(obj, source, **kwargs)

            for value in result:
                if isinstance(value, Model):
                    value = (value.pk, value)

                if value is not None:
                    if not isinstance(value, (tuple, list)):
                        value = (value, value)
                    values.append(value)

        if len(values) == 1:
            value = values[0]
            if value is None and self.empty_value is not None:
                value = self.empty_value
        elif len(values) > 0:
            plain_value = [v[0] for v in values]
            rich_value = self.separator.join(map(six.text_type, [v[1] for v in values]))
            value = (plain_value, rich_value)
        else:
            value = self.empty_value

        return value

    def get_source_value(self, obj, source, **kwargs):
        """
        Retrieves the value from ``obj`` associated with some ``source`` description.  Should return
        a list whose length is determined by the number of sources consulted.  In the default case,
        this is always just 1, but compound columns that declare their components with nested
        ``Column`` instances will have sources of their own and need to return a value per nested
        source.
        """
        if hasattr(source, "__call__"):
            value = source(obj)
        elif isinstance(obj, Model):
            value = reduce(get_attribute_value, [obj] + source.split('__'))
        elif isinstance(obj, dict):  # ValuesQuerySet item
            value = obj[source]
        else:
            raise ValueError("Unknown object type %r" % (repr(obj),))
        return [value]

    def get_processor_kwargs(self, **extra_kwargs):
        """
        Returns a dictionary of kwargs that should be sent to this column's :py:attr:`processor`
        callback.
        """
        kwargs = {
            'localize': self.localize,
        }
        kwargs.update(extra_kwargs)
        return kwargs

    def get_db_sources(self, model):
        """
        Returns the list of sources that match fields on the given ``model`` class.
        """
        sources = []
        for source in self.sources:
            target_field = self.resolve_source(model, source)
            if target_field:
                sources.append(source)
        return sources

    def get_virtual_sources(self, model):
        """
        Returns the list of sources that do not match fields on the given ``model`` class.
        """
        sources = []
        for source in self.sources:
            target_field = self.resolve_source(model, source)
            if target_field is None:
                sources.append(source)
        return sources

    def get_sort_fields(self, model):
        """
        Returns the list of sources that match fields on the given ``model`` class.  This will be
        the list of database-backed fields.
        """
        return self.get_db_sources(model)

    def expand_source(self, source):
        return (source,)

    def resolve_source(self, model, source):
        # Try to fetch the leaf attribute.  If this fails, the attribute is not database-backed and
        # the search for the first non-database field should end.
        if hasattr(source, "__call__"):
            return None
        try:
            return resolve_orm_path(model, source)
        except FieldDoesNotExist:
            return None

    def get_source_handler(self, model, source):
        """ Return handler instance for lookup types and term coercion. """
        return self

    # Interactivity features
    def prep_search_value(self, term, lookup_type):
        """
        Coerce the input term to work for the given lookup_type.  Returns the coerced term, or
        ``None`` if the term and lookup_type are incompatible together.
        """

        # We avoid making changes that the Django ORM can already do for us
        multi_terms = None

        if isinstance(term, six.text_type):
            if lookup_type == "in":
                in_bits = re.split(r',\s*', term)
                if len(in_bits) > 1:
                    multi_terms = in_bits
                else:
                    term = None

            if lookup_type == "range":
                range_bits = re.split(r'\s*-\s*', term)
                if len(range_bits) == 2:
                    multi_terms = range_bits
                else:
                    term = None

        if multi_terms:
            return filter(None, (self.prep_search_value(multi_term, lookup_type) for multi_term in multi_terms))

        model_field = self.model_field_class()
        try:
            term = model_field.get_prep_value(term)
        except:
            term = None

        return term

    def get_lookup_types(self, handler=None):
        """
        Generates the list of valid ORM lookup operators, taking into account runtime options for
        the ``allow_regex`` and ``allow_full_text_search`` options.
        """
        lookup_types = self.lookup_types
        if handler:
            lookup_types = handler.lookup_types

        # Add regex and MySQL 'search' operators if requested for the original column definition
        if self.allow_regex and 'iregex' not in lookup_types:
            lookup_types += ('iregex',)
        if self.allow_full_text_search and 'search' not in lookup_types:
            lookup_types += ('search',)
        return lookup_types

    def search(self, model, term, lookup_types=None):
        """
        Returns the ``Q`` object representing queries to make against this column for the given
        term.

        It is the responsibility of this method to decide which of this column's sources are
        database-backed and which are names of instance attributes, properties, or methods.  The
        ``model`` is provided to identify Django ORM ``ModelField`` s and related models.

        The default implementation resolves each contributing ``source`` from :py:attr:`sources`,
        and uses :py:meth:`.prep_search_value` to coerce the input search ``term`` to something
        usable for each of the query :py:attr:`lookup_types` supported by the column.  Any failed
        coercions will be skipped.

        The default implementation will also discover terms that match the source field's
        ``choices`` labels, flipping the term to automatically query for the internal choice value.
        """
        sources = self.get_db_sources(model)
        column_queries = []
        for source in sources:
            handler = self.get_source_handler(model, source)

            for sub_source in self.expand_source(source):
                modelfield = resolve_orm_path(model, sub_source)
                if modelfield.choices:
                    if hasattr(modelfield, 'get_choices'):
                        choices = modelfield.get_choices()
                    else:
                        choices = modelfield.get_flatchoices()
                    for db_value, label in choices:
                        if term.lower() in label.lower():
                            k = '%s__exact' % (sub_source,)
                            column_queries.append(Q(**{k: str(db_value)}))

                if not lookup_types:
                    lookup_types = handler.get_lookup_types()
                for lookup_type in lookup_types:
                    coerced_term = handler.prep_search_value(term, lookup_type)
                    if coerced_term is None:
                        # Skip terms that don't work with the lookup_type
                        continue
                    elif lookup_type in ('in', 'range') and not isinstance(coerced_term, tuple):
                        # Skip attempts to build multi-component searches if we only have one term
                        continue

                    k = '%s__%s' % (sub_source, lookup_type)
                    column_queries.append(Q(**{k: coerced_term}))

        if column_queries:
            q = reduce(operator.or_, column_queries)
        else:
            q = None
        return q

    # Template rendering
    def __str__(self):
        """
        Renders a simple ``<th>`` element with ``data-name`` attribute.  All items found in the
        ``self.attributes`` dict are also added as dom attributes.
        """
        return mark_safe(u"""<th data-name="{name_slug}"{attrs}>{label}</th>""".format(**{
            'name_slug': slugify(self.label),
            'attrs': self.attributes,
            'label': self.label,
        }))

    @property
    def attributes(self):
        """
        Returns a dictionary of initial state data for sorting, sort direction, and visibility.

        The default attributes include ``data-config-sortable``, ``data-config-visible``, and (if
        applicable) ``data-config-sorting`` to hold information about the initial sorting state.
        """
        attributes = {
            'data-config-sortable': 'true' if self.sortable else 'false',
            'data-config-visible': 'true' if self.visible else 'false',
        }

        if self.sort_priority is not None:
            attributes['data-config-sorting'] = ','.join(map(six.text_type, [
                self.sort_priority,
                self.index,
                self.sort_direction,
            ]))

        return flatatt(attributes)


class TextColumn(Column):
    model_field_class = models.CharField
    handles_field_classes = [models.CharField, models.TextField, models.FileField]

    # Add UUIDField if present in this version of Django
    try:
        handles_field_classes.append(models.UUIDField)
    except AttributeError:
        pass

    lookup_types = ('icontains', 'in')


class DateColumn(Column):
    model_field_class = models.DateField
    handles_field_classes = [models.DateField]
    lookup_types = ('exact', 'in', 'range', 'year', 'month', 'day', 'week_day')

    def prep_search_value(self, term, lookup_type):
        if lookup_type in ('exact', 'in', 'range'):
            try:
                date_obj = dateutil.parser.parse(term)
            except ValueError:
                # This exception is theoretical, but it doesn't seem to raise.
                pass
            except TypeError:
                # Failed conversions can lead to the parser adding ints to None.
                pass
            else:
                return date_obj

        if lookup_type not in ('exact', 'in', 'range'):
            test_term = term
            if lookup_type == 'week_day':
                try:
                    test_term = int(test_term) - 1  # Django ORM uses 1-7, python strptime uses 0-6
                except:
                    return None
                else:
                    test_term = str(test_term)

            for test_format in STRPTIME_PLACEHOLDERS[lookup_type]:
                # Try to validate the term against the given date lookup type
                try:
                    date_obj = datetime.strptime(test_term, test_format)
                except ValueError:
                    pass
                else:
                    if lookup_type == 'week_day':
                        term = date_obj.weekday() + 1  # Django ORM uses 1-7, python strptime uses 0-6
                    else:
                        term = getattr(date_obj, lookup_type)
                    return str(term)

        return super(DateColumn, self).prep_search_value(term, lookup_type)


class DateTimeColumn(DateColumn):
    model_field_class = models.DateTimeField
    handles_field_classes = [models.DateTimeField]
    lookup_types = ('exact', 'in', 'range', 'year', 'month', 'day', 'week_day', 'hour', 'minute', 'second')


class TimeColumn(DateColumn):
    model_field_class = models.TimeField
    handles_field_classes = [models.TimeField]
    lookup_types = ('exact', 'in', 'range', 'hour', 'minute', 'second')


class BooleanColumn(Column):
    model_field_class = models.BooleanField
    handles_field_classes = [models.BooleanField, models.NullBooleanField]
    lookup_types = ('exact', 'in')

    def prep_search_value(self, term, lookup_type):
        term = term.lower()
        # Allow column's own label to represent a true value
        if term == 'true' or term.lower() in self.label.lower():
            term = True
        elif term == 'false':
            term = False
        else:
            return None
        return super(BooleanColumn, self).prep_search_value(term, lookup_type)


class IntegerColumn(Column):
    model_field_class = models.IntegerField
    handles_field_classes = [models.IntegerField, models.AutoField]
    lookup_types = ('exact', 'in')


class FloatColumn(Column):
    model_field_class = models.FloatField
    handles_field_classes = [models.FloatField, models.DecimalField]
    lookup_types = ('exact', 'in')


class CompoundColumn(Column):
    """
    Special column type for holding multiple sources that have different model field types.  The
    separation of sources by type is important because of the different query lookup types that are
    allowed against different model fields.

    Each source will dynamically find its associated model field and choose an appropriate column
    class from the registry.

    To more finely control which column class is used, an actual column instance can be given
    instead of a string name which declares its own ``source`` or ``sources``.  Because they are not
    important to the client-side representation of the column, no ``label`` is necessary for nested
    column instances.
    """

    model_field_class = None
    handles_field_classes = []
    lookup_types = ()

    def expand_source(self, source):
        if isinstance(source, Column):
            return source.sources
        return super(CompoundColumn, self).expand_source(source)

    def get_source_value(self, obj, source, **kwargs):
        """
        Treat ``field`` as a nested sub-Column instance, which explicitly stands in as the object
        to which term coercions and the query type lookup are delegated.
        """
        result = []
        for sub_source in self.expand_source(source):
            # Call super() to get default logic, but send it the 'sub_source'
            sub_result = super(CompoundColumn, self).get_source_value(obj, sub_source, **kwargs)
            result.extend(sub_result)
        return result

    def get_db_sources(self, model):
        return self.sources

    def get_sort_fields(self, model):
        return self._get_flat_db_sources(model)

    def _get_flat_db_sources(self, model):
        """ Return a flattened representation of the individual ``sources`` lists. """
        sources = []
        for source in self.sources:
            for sub_source in self.expand_source(source):
                target_field = self.resolve_source(model, sub_source)
                if target_field:
                    sources.append(sub_source)
        return sources

    def get_source_handler(self, model, source):
        """ Allow the nested Column source to be its own handler. """
        if isinstance(source, Column):
            return source

        # Generate a generic handler for the source
        modelfield = resolve_orm_path(model, source)
        column_class = get_column_for_modelfield(modelfield)
        return column_class()


class DisplayColumn(Column):
    """
    Convenience column type for unsearchable, unsortable columns, which rely solely on a processor
    function to generate content.
    """

    model_field_class = None
    lookup_types = ()



# -*- encoding: utf-8 -*-
from .testcase import DatatableViewTestCase
from .test_app import models
from ..columns import Column, COLUMN_CLASSES
from ..exceptions import ColumnError
from .. import utils

class ColumnTests(DatatableViewTestCase):
    def test_custom_column_registers_itself(self):
        previous_length = len(COLUMN_CLASSES)

        class CustomColumn(Column):
            model_field_class = 'fake'

        self.assertEqual(len(COLUMN_CLASSES), previous_length + 1)
        self.assertEqual(COLUMN_CLASSES[0][0], CustomColumn)
        self.assertEqual(COLUMN_CLASSES[0][1], [CustomColumn.model_field_class] + CustomColumn.handles_field_classes)

        del COLUMN_CLASSES[:1]

    def test_value_is_pair(self):
        obj = models.ExampleModel.objects.create(name="test name 1")

        column = Column()
        value = column.value(obj)
        self.assertEqual(type(value), tuple)

    # def test_process_value_checks_all_sources(self):
    def test_process_value_is_empty_for_fake_source(self):
        processed = []

        def processor(value, **kwargs):
            processed.append(value)

        obj = models.ExampleModel.objects.create(name="test name 1")

        # Verify bad source names don't find values
        processed[:] = []
        column = Column(sources=['fake1'], processor=processor)
        column.value(obj)
        self.assertEqual(processed, [])

        column = Column(sources=['fake1', 'fake2'], processor=processor)
        column.value(obj)
        self.assertEqual(processed, [])



# -*- encoding: utf-8 -*-

from django import template
from django import get_version
from django.urls import reverse

register = template.Library()

if get_version().split('.') < ['1', '5']:

    @register.simple_tag(name="url")
    def django_1_4_url_simple(url_name):
        return reverse(url_name)
